__author__ = 'Robbert Harms'
__date__ = "2015-08-25"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ModelProtocolProblem(object):

    def __init__(self):
        """The base class for indicating problems with a protocol.

        These are meant to be returned from the function get_protocol_problems() from the ProtocolCheckInterface.

        Each of these problems is supposed to overwrite the function __str__() for reporting the problem.
        """

    def __repr__(self):
        return self.__str__()


class MissingColumns(ModelProtocolProblem):

    def __init__(self, missing_columns):
        super(MissingColumns, self).__init__()
        self.missing_columns = missing_columns

    def __str__(self):
        return 'Missing columns: ' + ', '.join(self.missing_columns)


class InsufficientShells(ModelProtocolProblem):

    def __init__(self, required_nmr_shells, nmr_shells):
        super(InsufficientShells, self).__init__()
        self.required_nmr_shells = required_nmr_shells
        self.nmr_shells = nmr_shells

    def __str__(self):
        return 'Required number of shells is {}, this protocol has {}.'.format(
            self.required_nmr_shells, self.nmr_shells)


class NamedProtocolProblem(ModelProtocolProblem):

    def __init__(self, model_protocol_problem, model_name):
        """This extends the given model protocol problem to also include the name of the model.

        Args:
            model_protocol_problem (ModelProtocolProblem): The name for the problem with the given model.
            model_name (str): the name of the model
        """
        super(NamedProtocolProblem, self).__init__()
        self._model_protocol_problem = model_protocol_problem
        self._model_name = model_name

    def __str__(self):
        return "{0}: {1}".format(self._model_name, self._model_protocol_problem)



import glob
import os
import mdt
from mdt.batch_utils import SimpleBatchProfile, SimpleSubjectInfo

__author__ = 'Robbert Harms'
__date__ = "2015-07-13"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"

meta_info = {'title': 'HCP WU-Minn',
             'description': 'The profile for the WU-Minn data from the Human Connectome project',
             'directory_layout':
'''
This assumes that you downloaded and extracted the WU-Minn data in one folder which gives one folder per subject.

You can provide the noise standard deviation to use using a noise_std file containing a single float.

Example directory layout:
    /*/T1w/Diffusion/data.nii.gz
    /*/T1w/Diffusion/bvals
    /*/T1w/Diffusion/bvecs
    /*/T1w/Diffusion/nodif_brain_mask.nii.gz
    /*/T1w/Diffusion/grad_dev.nii.gz

Optional items (these will take precedence if present):
    /*/T1w/Diffusion/data.bval
    /*/T1w/Diffusion/data.bvec
    /*/T1w/Diffusion/data.prtcl
    /*/T1w/Diffusion/data_mask.nii(.gz)
    /*/T1w/Diffusion/noise_std
'''}


class HCP_WUMINN(SimpleBatchProfile):

    def __init__(self):
        super(HCP_WUMINN, self).__init__()
        self.use_gradient_deviations = False
        self._output_base_dir = 'T1w/Diffusion/output'
        self._append_mask_name_to_output_sub_dir = False

    def _get_subjects(self):
        subjects = []
        for subject_id in sorted([os.path.basename(f) for f in glob.glob(os.path.join(self._root_dir, '*'))]):
            pjoin = mdt.make_path_joiner(self._root_dir, subject_id, 'T1w', 'Diffusion')
            if os.path.isdir(pjoin()):
                subject_info = self._get_subject_in_directory(subject_id, pjoin)
                if subject_info:
                    subjects.append(subject_info)
        return subjects

    def _get_subject_in_directory(self, subject_id, pjoin):
        """Get the information about the given subject in the given directory.

        Args:
            subject_id (str): the id of this subject we are loading
            pjoin (PathJoiner): the path joiner pointing to the directory of that subject

        Returns:
            SimpleSubjectInfo or None: the subject info for this particular subject
        """
        noise_std = self._autoload_noise_std(subject_id, file_path=pjoin('noise_std'))

        protocol_loader = self._autoload_protocol(
            pjoin(),
            protocols_to_try=['data.prtcl'],
            bvals_to_try=['data.bval', 'bvals'],
            bvecs_to_try=['data.bvec', 'bvecs'],
            protocol_columns={'Delta': 43.1e-3, 'delta': 10.6e-3, 'TE': 89.5e-3, 'TR': 5520e-3})

        mask_fname = self._get_first_existing_nifti(['data_mask', 'nodif_brain_mask'], prepend_path=pjoin())

        output_dir = self._get_subject_output_dir(subject_id, mask_fname)

        return SimpleSubjectInfo(subject_id, pjoin('data'), protocol_loader, mask_fname,
                                 output_dir, gradient_deviations=pjoin('grad_dev'),
                                 use_gradient_deviations=self.use_gradient_deviations, noise_std=noise_std)

    def __str__(self):
        return meta_info['title']



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
"""Print some basic information about an image file."""
import argparse
import os
from mdt.nifti import load_nifti
import textwrap

from mdt.shell_utils import BasicShellApplication

__author__ = 'Robbert Harms'
__date__ = "2015-08-18"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class InfoImg(BasicShellApplication):

    def _get_arg_parser(self):
        description = textwrap.dedent(__doc__)
        description += self._get_citation_message()

        epilog = textwrap.dedent("""
            Examples of use:
                mdt-info-img my_img.nii
                mdt-info-img *.nii
        """)

        parser = argparse.ArgumentParser(description=description, epilog=epilog,
                                         formatter_class=argparse.RawTextHelpFormatter)

        parser.add_argument('images', metavar='images', nargs="+", type=str,
                            help="The input images")

        return parser

    def run(self, args):
        for image in args.images:
            image_path = os.path.realpath(image)
            img = load_nifti(image_path)
            header = img.get_header()
            print('{}'.format(image))
            self.print_info(header)
            print('')

    def print_info(self, header):
        row_format = "{:<15}{}"
        print(row_format.format('data_type', str(header.get_data_dtype()).upper()))
        print(row_format.format('nmr_dim', len(header.get_data_shape())))

        for ind, el in enumerate(header.get_data_shape()):
            print(row_format.format('dim{}'.format(ind), el))

        for ind, el in enumerate(header.get_zooms()):
            print(row_format.format('pixdim{}'.format(ind), el))


if __name__ == '__main__':
    InfoImg().start()



import numpy as np

import mdt
from mdt.components_loader import bind_function
from mdt.models.compartments import CompartmentConfig
from mdt.cl_routines.mapping.dti_measures import DTIMeasures
from mdt.utils import eigen_vectors_from_tensor

__author__ = 'Robbert Harms'
__date__ = "2015-06-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class Tensor(CompartmentConfig):

    parameter_list = ('g', 'b', 'd', 'dperp0', 'dperp1', 'theta', 'phi', 'psi')

    @bind_function
    def get_extra_results_maps(self, results_dict):
        eigen_vectors = eigen_vectors_from_tensor(results_dict[self.name + '.theta'], results_dict[self.name + '.phi'],
                                                  results_dict[self.name + '.psi'])

        eigen_values = Tensor.create_eigen_values_matrix([results_dict[self.name + '.d'],
                                                          results_dict[self.name + '.dperp0'],
                                                          results_dict[self.name + '.dperp1']])

        ranking = Tensor.get_ranking_matrix(eigen_values)

        voxels_range = np.arange(ranking.shape[0])
        sorted_eigen_values = np.concatenate([eigen_values[voxels_range, ranking[:, ind], None]
                                              for ind in range(ranking.shape[1])], axis=1)
        sorted_eigen_vectors = np.concatenate([eigen_vectors[voxels_range, ranking[:, ind], None, :]
                                               for ind in range(ranking.shape[1])], axis=1)

        fa, md = DTIMeasures().calculate(eigen_values)

        extra_maps = {self.name + '.eigen_ranking': ranking,
                      self.name + '.FA': fa,
                      self.name + '.MD': md,
                      self.name + '.AD': sorted_eigen_values[:, 0],
                      self.name + '.RD': (sorted_eigen_values[:, 1] + sorted_eigen_values[:, 2]) / 2.0}

        for ind in range(3):
            extra_maps.update({self.name + '.vec' + repr(ind): eigen_vectors[:, ind, :],
                               self.name + '.sorted_vec' + repr(ind): sorted_eigen_vectors[:, ind, :],
                               self.name + '.sorted_eigval{}'.format(ind): sorted_eigen_values[:, ind]})

            for dimension in range(3):
                extra_maps.update({self.name + '.vec' + repr(ind) + '_' + repr(dimension):
                                       eigen_vectors[:, ind, dimension],
                                   self.name + '.sorted_vec' + repr(ind) + '_' + repr(dimension):
                                       sorted_eigen_vectors[:, ind, dimension]
                                   })

        return extra_maps

    @staticmethod
    def ensure_2d(array):
        if len(array.shape) < 2:
            return array[None, :]
        return array

    @staticmethod
    def create_eigen_values_matrix(diffusivities):
        return Tensor.ensure_2d(np.squeeze(np.dstack(diffusivities)))

    @staticmethod
    def get_ranking_matrix(eigen_values):
        return Tensor.ensure_2d(np.squeeze(np.argsort(eigen_values, axis=1)[:, ::-1]))



import os

from mdt.nifti import load_nifti
import numpy as np
from PyQt5.QtCore import pyqtSlot, QObject, pyqtSignal
from PyQt5.QtWidgets import QFileDialog

from mdt import load_brain_mask, create_median_otsu_brain_mask
from mdt.visualization.maps.base import DataInfo, MapPlotConfig
from mdt.gui.maps_visualizer.main import start_gui
from mdt.gui.model_fit.design.ui_generate_brain_mask_tab import Ui_GenerateBrainMaskTabContent
from mdt.gui.utils import function_message_decorator, image_files_filters, protocol_files_filters, MainTab

__author__ = 'Robbert Harms'
__date__ = "2016-06-26"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class GenerateBrainMaskTab(MainTab, Ui_GenerateBrainMaskTabContent):

    def __init__(self, shared_state, computations_thread):
        self._shared_state = shared_state
        self._computations_thread = computations_thread
        self._generate_mask_worker = GenerateMaskWorker()

    def setupUi(self, tab_content):
        super(GenerateBrainMaskTab, self).setupUi(tab_content)

        self.selectImageButton.clicked.connect(lambda: self._select_image())
        self.selectProtocolButton.clicked.connect(lambda: self._select_protocol())
        self.selectOutputButton.clicked.connect(lambda: self._select_output())
        self.viewButton.clicked.connect(self.view_mask)
        self.generateButton.clicked.connect(self.generate_mask)

        self.selectedImageText.textChanged.connect(self._check_enable_action_buttons)
        self.selectedOutputText.textChanged.connect(self._check_enable_action_buttons)
        self.selectedProtocolText.textChanged.connect(self._check_enable_action_buttons)

    def _select_image(self):
        initial_dir = self._shared_state.base_dir
        if self.selectedImageText.text() != '':
            initial_dir = self.selectedImageText.text()

        open_file, used_filter = QFileDialog().getOpenFileName(
            caption='Select the 4d diffusion weighted image', directory=initial_dir,
            filter=';;'.join(image_files_filters))

        if os.path.isfile(open_file):
            self.selectedImageText.setText(open_file)
            self._shared_state.base_dir = os.path.dirname(open_file)

    def _select_protocol(self):
        initial_dir = self._shared_state.base_dir
        if self.selectedProtocolText.text() != '':
            initial_dir = self.selectedProtocolText.text()

        open_file, used_filter = QFileDialog().getOpenFileName(
            caption='Select the protocol', directory=initial_dir,
            filter=';;'.join(protocol_files_filters))

        if os.path.isfile(open_file):
            self.selectedProtocolText.setText(open_file)
            self._shared_state.base_dir = os.path.dirname(open_file)

    def _select_output(self):
        initial_dir = self._shared_state.base_dir
        if self.selectedOutputText.text() != '':
            initial_dir = self.selectedOutputText.text()

        output_file_name, used_filter = QFileDialog().getSaveFileName(
            caption='Select the output file', directory=initial_dir,
            filter=';;'.join(image_files_filters))

        if output_file_name:
            self.selectedOutputText.setText(output_file_name)

    def _check_enable_action_buttons(self):
        self.generateButton.setEnabled(os.path.isfile(self.selectedImageText.text()) and
                                       os.path.isfile(self.selectedProtocolText.text())
                                       and os.path.isdir(os.path.dirname(self.selectedOutputText.text())))
        self.viewButton.setEnabled(os.path.isfile(self.selectedImageText.text()) and
                                   os.path.isfile(self.selectedOutputText.text()))

    def view_mask(self):
        mask = np.expand_dims(load_brain_mask(self.selectedOutputText.text()), axis=3)
        image_data = load_nifti(self.selectedImageText.text()).get_data()
        masked_image = image_data * mask

        data = DataInfo({'Masked': masked_image, 'DWI': image_data},
                        directory=os.path.dirname(self.selectedImageText.text()))

        config = MapPlotConfig()
        config.dimension = 2
        config.slice_index = image_data.shape[2] // 2.0
        config.maps_to_show = ['DWI', 'Masked']

        start_gui(data=data, config=config, app_exec=False)

    def generate_mask(self):
        self._generate_mask_worker.set_args(self.selectedImageText.text(),
                                            self.selectedProtocolText.text(),
                                            self.selectedOutputText.text(),
                                            median_radius=self.medianRadiusInput.value(),
                                            numpass=self.numberOfPassesInput.value(),
                                            mask_threshold=self.finalThresholdInput.value())
        self._computations_thread.start()
        self._generate_mask_worker.moveToThread(self._computations_thread)

        self._generate_mask_worker.starting.connect(self._computations_thread.starting)
        self._generate_mask_worker.finished.connect(self._computations_thread.finished)

        self._generate_mask_worker.starting.connect(lambda: self.generateButton.setEnabled(False))
        self._generate_mask_worker.finished.connect(lambda: self.generateButton.setEnabled(True))
        self._generate_mask_worker.finished.connect(lambda: self.viewButton.setEnabled(True))

        self._generate_mask_worker.starting.emit()


class GenerateMaskWorker(QObject):

    starting = pyqtSignal()
    finished = pyqtSignal()

    def __init__(self):
        super(GenerateMaskWorker, self).__init__()
        self.starting.connect(self.run)
        self._args = []
        self._kwargs = {}

    def set_args(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    @function_message_decorator('Started creating a mask.', 'Finished creating a mask.')
    def run(self):
        create_median_otsu_brain_mask(*self._args, **self._kwargs)
        self.finished.emit()



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
"""Evaluate an expression on a protocol.

This is meant to quickly change a protocol using mathematical expressions. The expressions
can be any valid python string separated if needed with the semicolon (;).

The columns of the input protocol are loaded and stored as arrays with as variable names the names of the
columns. Next, the expression is evaluated on those columns and the result is stored in the indicated file.

An additional function "rm(<column_name>)" is also available with wich you can remove columns from
the protocol, and a function "add(<column_name>, <value>)" is available to add columns. When adding
a column, the value can either be a scalar or a vector.

Additionally the numpy library is available with prefix 'np.'.
"""
import argparse
import os
import numpy as np
import mdt
from argcomplete.completers import FilesCompleter
import textwrap
from mdt.shell_utils import BasicShellApplication

__author__ = 'Robbert Harms'
__date__ = "2015-08-18"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class MathProtocol(BasicShellApplication):

    def _get_arg_parser(self):
        description = textwrap.dedent(__doc__)
        description += self._get_citation_message()

        epilog = textwrap.dedent("""
            Examples of use:
                mdt-math-protocol protocol.prtcl 'G *= 1e-3'
                mdt-math-protocol protocol.prtcl 'G *= 1e-3; TR /= 1000; TE /= 1000'
                mdt-math-protocol protocol.prtcl "rm('G')"
                mdt-math-protocol protocol.prtcl "add('TE', 50e-3)"
        """)

        parser = argparse.ArgumentParser(description=description, epilog=epilog,
                                         formatter_class=argparse.RawTextHelpFormatter)

        parser.add_argument('input_protocol', metavar='input_protocol', type=str,
                            help="The input protocol")

        parser.add_argument('expr', metavar='expr', type=str,
                            help="The expression to evaluate.")

        parser.add_argument('-o', '--output_file',
                            help='the output protocol, defaults to the input protocol.').completer = FilesCompleter()

        return parser

    def run(self, args):
        if args.output_file is not None:
            output_file = os.path.realpath(args.output_file)
        else:
            output_file = os.path.realpath(args.input_protocol)

        protocol = mdt.load_protocol(os.path.realpath(args.input_protocol))
        context_dict = {name: protocol.get_column(name) for name in protocol.column_names}

        def rm(column_name):
            protocol.remove_column(column_name)
            del context_dict[column_name]

        def add(column_name, value):
            protocol.add_column(column_name, value)
            context_dict[column_name] = value

        exec(args.expr, {'np': np, 'rm': rm, 'add': add}, context_dict)

        for name, value in context_dict.items():
            protocol.update_column(name, value)

        mdt.write_protocol(protocol, output_file)


if __name__ == '__main__':
    MathProtocol().start()



from mdt.models.compartments import CompartmentConfig

__author__ = 'Robbert Harms'
__date__ = "2015-06-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class SphereGPD(CompartmentConfig):

    parameter_list = ('Delta', 'delta', 'd', 'R')
    dependency_list = ('MRIConstants',)



import copy
from mdt.visualization.dict_conversion import ConvertDictElements
from mdt.visualization.maps.base import SingleMapConfig, MapPlotConfig, Zoom, Point


class PlottingFrame(object):

    def __init__(self, controller, plotting_info_viewer=None):
        super(PlottingFrame, self).__init__()
        self._controller = controller
        self._plotting_info_viewer = plotting_info_viewer or NoOptPlottingFrameInfoViewer()

    def set_auto_rendering(self, auto_render):
        """Set if this plotting frame should auto render itself on every configuration update, or not.

        Args:
            auto_render (boolean): if True the plotting frame should auto render, if False it should only
                render on manual updates.
        """

    def redraw(self):
        """Tell the plotting frame to do a redraw."""

    def export_image(self, filename, width, height, dpi=100):
        """Export the current view as an image.

        Args:
            filename (str): where to write the file
            width (int): the width in pixels
            height (int): the height in pixels
            dpi (int): the dpi of the result
        """


class PlottingFrameInfoViewer(object):

    def __init__(self):
        """Implementations of this class can be given to a PlottingFrame to update viewing information.

        As an interface is bridges the gap between the rest of the GUI and the PlottingFrame and
        can encapsulate highlighting interesting aspects of one of the plots.
        """

    def set_voxel_info(self, onscreen_coords, data_index, value):
        """Highlight a single voxel.

        Args:
            onscreen_coords (tuple of x,y): the coordinates of the voxel onscreen
            data_index (tuple of x,y,z,v): the 4d coordinates of the corresponding voxel in the data
            value (float): the value of the object in the 4d coordinates.
        """

    def clear_voxel_info(self):
        """Tell the info viewer that we are no longer looking at a specific voxel."""


class NoOptPlottingFrameInfoViewer(PlottingFrameInfoViewer):

    def set_voxel_info(self, onscreen_coords, data_index, value):
        super(NoOptPlottingFrameInfoViewer, self).set_voxel_info(onscreen_coords, data_index, value)
        print(onscreen_coords, data_index, value)

    def clear_voxel_info(self):
        super(NoOptPlottingFrameInfoViewer, self).clear_voxel_info()
        print('clear')


def cast_value(value, desired_type, alt_value):
    """Cast the given value to the desired type, on failure returns the alternative value.

    Args:
        value (object): the value to cast to the given type
        desired_type (:class:`type`): the type to cast to
        alt_value (object): the alternative value if casting threw exceptions

    Returns:
        the desired casted value or the alternative value if casting failed.
    """
    try:
        return desired_type(value)
    except TypeError:
        return alt_value
    except ValueError:
        return alt_value


class ConfigAction(object):

    def __init__(self):
        """Allows apply and unapply of configuration changes."""
        self._previous_config = None

    def apply(self, data_info, configuration):
        """Apply the changes to the given configuration and return a new one.

        This should return a new configuration with the applied changes and should not update the given configuration.

        By default this method calls _apply(configuration) to facilitate quick implementation.

        Args:
            data_info (DataInfo): the current data information
            configuration (DisplayConfiguration): the configuration object

        Returns:
            MapPlotConfig: the updated configuration
        """
        self._previous_config = configuration
        new_config = copy.deepcopy(configuration)
        updated_new_config = self._apply(data_info, new_config)
        if updated_new_config:
            return updated_new_config
        return new_config

    def unapply(self):
        """Return the configuration as it was before the application of this function.

        Returns:
            MapPlotConfig: the previous configuration
        """
        return self._previous_config

    def _apply(self, data_info, configuration):
        """Facilitates quick implementation, called by apply()

        One can set configuration changes immediately to the given configuration. If nothing is returned we
        will use the given configuration as the new configuration.

        Args:
            data_info (DataInfo): the current data information
            configuration (MapPlotConfig): the configuration object

        Returns:
            GeneralConfiguration or None: the updated configuration. If nothing is returned we use the one given as
                argument.
        """


class SimpleConfigAction(ConfigAction):

    config_attribute = None

    def __init__(self, new_value):
        """A simple configuration action this sets the given value to the config attribute of the configuration."""
        super(SimpleConfigAction, self).__init__()
        self.new_value = new_value

    def _apply(self, data_info, configuration):
        setattr(configuration, self.config_attribute, self.new_value)
        return self._extra_actions(data_info, configuration)

    def _extra_actions(self, data_info, configuration):
        """Called by the default configuration action to apply additional changes"""
        return configuration


class SimpleMapSpecificConfigAction(SimpleConfigAction):

    config_attribute = None

    def __init__(self, map_name, new_value):
        super(SimpleMapSpecificConfigAction, self).__init__(new_value)
        self.map_name = map_name

    def _apply(self, data_info, configuration):
        if self.map_name not in configuration.map_plot_options:
            configuration.map_plot_options[self.map_name] = SingleMapConfig()
        single_map_config = super(SimpleMapSpecificConfigAction, self)._apply(
            data_info,
            configuration.map_plot_options[self.map_name])

        if single_map_config is None:
            del configuration.map_plot_options[self.map_name]

        return configuration

    def _extra_actions(self, data_info, configuration):
        single_map_config = super(SimpleMapSpecificConfigAction, self)._extra_actions(data_info, configuration)
        if single_map_config == SingleMapConfig():
            return None
        return single_map_config


class Controller(object):

    def __init__(self):
        """Controller interface"""
        super(Controller, self).__init__()

    def set_data(self, data_info, config=None):
        """Set new data to visualize.

        Args:
            data_info (mdt.visualization.maps.base.DataInfo): the new data to visualize
            config (MapPlotConfig): the new configuration for the data
                If given, we will display the new data immediately with the given config
        """

    def get_data(self):
        """Get the current data.

        Returns:
            mdt.visualization.maps.base.DataInfo: the current data information
        """

    def set_config(self, general_config):
        """Set the general configuration to the given config.

        Setting this should automatically update all the listeners.

        Args:
            general_config (MapPlotConfig): the general configuration
        """

    def get_config(self):
        """Get the current configuration.

        Returns:
            MapPlotConfig: the current general configuration.
        """

    def apply_action(self, action):
        """Apply a new configuration action.

        If there is no difference between the current config and the one generated by this new action, the
        action will not be stored in history and will not need to be applied.

        Args:
            action (mdt.gui.maps_visualizer.base.ConfigAction): the configuration action to add and apply
        """

    def undo(self):
        """Undo a previous configuration action"""

    def redo(self):
        """Reapply a previously undone configuration action"""

    def has_undo(self):
        """Check if this controller has an undo action available.

        Returns:
            boolean: True if an undo action is available.
        """

    def has_redo(self):
        """Check if this controller has an redo action available.

        Returns:
            boolean: True if an redo action is available.
        """



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'generate_roi_mask_tab.ui'
#
# Created by: PyQt5 UI code generator 5.4.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_GenerateROIMaskTabContent(object):
    def setupUi(self, GenerateROIMaskTabContent):
        GenerateROIMaskTabContent.setObjectName("GenerateROIMaskTabContent")
        GenerateROIMaskTabContent.resize(827, 427)
        self.verticalLayout = QtWidgets.QVBoxLayout(GenerateROIMaskTabContent)
        self.verticalLayout.setContentsMargins(-1, 11, -1, -1)
        self.verticalLayout.setObjectName("verticalLayout")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setContentsMargins(-1, -1, -1, 0)
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label = QtWidgets.QLabel(GenerateROIMaskTabContent)
        font = QtGui.QFont()
        font.setPointSize(14)
        self.label.setFont(font)
        self.label.setObjectName("label")
        self.verticalLayout_2.addWidget(self.label)
        self.label_2 = QtWidgets.QLabel(GenerateROIMaskTabContent)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_2.setFont(font)
        self.label_2.setObjectName("label_2")
        self.verticalLayout_2.addWidget(self.label_2)
        self.verticalLayout.addLayout(self.verticalLayout_2)
        self.line = QtWidgets.QFrame(GenerateROIMaskTabContent)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setLineWidth(1)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setObjectName("line")
        self.verticalLayout.addWidget(self.line)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.gridLayout.setHorizontalSpacing(10)
        self.gridLayout.setObjectName("gridLayout")
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.dimensionInput = QtWidgets.QSpinBox(GenerateROIMaskTabContent)
        self.dimensionInput.setMinimum(0)
        self.dimensionInput.setMaximum(2)
        self.dimensionInput.setProperty("value", 2)
        self.dimensionInput.setObjectName("dimensionInput")
        self.horizontalLayout_5.addWidget(self.dimensionInput)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_5.addItem(spacerItem)
        self.gridLayout.addLayout(self.horizontalLayout_5, 1, 1, 1, 1)
        self.label_6 = QtWidgets.QLabel(GenerateROIMaskTabContent)
        self.label_6.setObjectName("label_6")
        self.gridLayout.addWidget(self.label_6, 0, 0, 1, 1)
        self.label_3 = QtWidgets.QLabel(GenerateROIMaskTabContent)
        self.label_3.setMinimumSize(QtCore.QSize(0, 0))
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 3, 0, 1, 1)
        self.label_10 = QtWidgets.QLabel(GenerateROIMaskTabContent)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_10.setFont(font)
        self.label_10.setObjectName("label_10")
        self.gridLayout.addWidget(self.label_10, 2, 2, 1, 1)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.selectMaskButton = QtWidgets.QPushButton(GenerateROIMaskTabContent)
        self.selectMaskButton.setObjectName("selectMaskButton")
        self.horizontalLayout_2.addWidget(self.selectMaskButton)
        self.selectedMaskText = QtWidgets.QLineEdit(GenerateROIMaskTabContent)
        self.selectedMaskText.setText("")
        self.selectedMaskText.setObjectName("selectedMaskText")
        self.horizontalLayout_2.addWidget(self.selectedMaskText)
        self.gridLayout.addLayout(self.horizontalLayout_2, 0, 1, 1, 1)
        self.label_4 = QtWidgets.QLabel(GenerateROIMaskTabContent)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_4.setFont(font)
        self.label_4.setObjectName("label_4")
        self.gridLayout.addWidget(self.label_4, 0, 2, 1, 1)
        self.label_5 = QtWidgets.QLabel(GenerateROIMaskTabContent)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_5.setFont(font)
        self.label_5.setObjectName("label_5")
        self.gridLayout.addWidget(self.label_5, 1, 2, 1, 1)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setContentsMargins(-1, 0, -1, -1)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.sliceInput = QtWidgets.QSpinBox(GenerateROIMaskTabContent)
        self.sliceInput.setMinimum(0)
        self.sliceInput.setMaximum(10000)
        self.sliceInput.setProperty("value", 0)
        self.sliceInput.setObjectName("sliceInput")
        self.horizontalLayout_3.addWidget(self.sliceInput)
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setContentsMargins(3, -1, -1, -1)
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.label_7 = QtWidgets.QLabel(GenerateROIMaskTabContent)
        self.label_7.setObjectName("label_7")
        self.horizontalLayout_6.addWidget(self.label_7)
        self.maxSliceLabel = QtWidgets.QLabel(GenerateROIMaskTabContent)
        self.maxSliceLabel.setObjectName("maxSliceLabel")
        self.horizontalLayout_6.addWidget(self.maxSliceLabel)
        self.horizontalLayout_3.addLayout(self.horizontalLayout_6)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem1)
        self.gridLayout.addLayout(self.horizontalLayout_3, 2, 1, 1, 1)
        self.label_11 = QtWidgets.QLabel(GenerateROIMaskTabContent)
        self.label_11.setObjectName("label_11")
        self.gridLayout.addWidget(self.label_11, 1, 0, 1, 1)
        self.label_14 = QtWidgets.QLabel(GenerateROIMaskTabContent)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_14.setFont(font)
        self.label_14.setObjectName("label_14")
        self.gridLayout.addWidget(self.label_14, 3, 2, 1, 1)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setContentsMargins(0, -1, 0, -1)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.selectOutputFileInput = QtWidgets.QPushButton(GenerateROIMaskTabContent)
        self.selectOutputFileInput.setObjectName("selectOutputFileInput")
        self.horizontalLayout_4.addWidget(self.selectOutputFileInput)
        self.selectedOutputFileText = QtWidgets.QLineEdit(GenerateROIMaskTabContent)
        self.selectedOutputFileText.setObjectName("selectedOutputFileText")
        self.horizontalLayout_4.addWidget(self.selectedOutputFileText)
        self.gridLayout.addLayout(self.horizontalLayout_4, 3, 1, 1, 1)
        self.label_12 = QtWidgets.QLabel(GenerateROIMaskTabContent)
        self.label_12.setObjectName("label_12")
        self.gridLayout.addWidget(self.label_12, 2, 0, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        self.line_2 = QtWidgets.QFrame(GenerateROIMaskTabContent)
        self.line_2.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.verticalLayout.addWidget(self.line_2)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setContentsMargins(-1, 6, -1, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.generateButton = QtWidgets.QPushButton(GenerateROIMaskTabContent)
        self.generateButton.setEnabled(False)
        self.generateButton.setObjectName("generateButton")
        self.horizontalLayout.addWidget(self.generateButton)
        self.viewButton = QtWidgets.QPushButton(GenerateROIMaskTabContent)
        self.viewButton.setEnabled(False)
        self.viewButton.setObjectName("viewButton")
        self.horizontalLayout.addWidget(self.viewButton)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem2)
        self.verticalLayout.addLayout(self.horizontalLayout)
        spacerItem3 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem3)

        self.retranslateUi(GenerateROIMaskTabContent)
        QtCore.QMetaObject.connectSlotsByName(GenerateROIMaskTabContent)
        GenerateROIMaskTabContent.setTabOrder(self.selectMaskButton, self.selectedMaskText)
        GenerateROIMaskTabContent.setTabOrder(self.selectedMaskText, self.dimensionInput)
        GenerateROIMaskTabContent.setTabOrder(self.dimensionInput, self.sliceInput)
        GenerateROIMaskTabContent.setTabOrder(self.sliceInput, self.selectOutputFileInput)
        GenerateROIMaskTabContent.setTabOrder(self.selectOutputFileInput, self.selectedOutputFileText)
        GenerateROIMaskTabContent.setTabOrder(self.selectedOutputFileText, self.generateButton)
        GenerateROIMaskTabContent.setTabOrder(self.generateButton, self.viewButton)

    def retranslateUi(self, GenerateROIMaskTabContent):
        _translate = QtCore.QCoreApplication.translate
        GenerateROIMaskTabContent.setWindowTitle(_translate("GenerateROIMaskTabContent", "Form"))
        self.label.setText(_translate("GenerateROIMaskTabContent", "Generate ROI mask"))
        self.label_2.setText(_translate("GenerateROIMaskTabContent", "Create a mask with a Region Of Interest including only the voxels in the selected slice."))
        self.label_6.setText(_translate("GenerateROIMaskTabContent", "Select brain mask:"))
        self.label_3.setText(_translate("GenerateROIMaskTabContent", "Select output file:"))
        self.label_10.setText(_translate("GenerateROIMaskTabContent", "(The index of the single slice in the current dimension)"))
        self.selectMaskButton.setText(_translate("GenerateROIMaskTabContent", "Browse"))
        self.label_4.setText(_translate("GenerateROIMaskTabContent", "(Select your brain mask)"))
        self.label_5.setText(_translate("GenerateROIMaskTabContent", "(The dimension of the single slice)"))
        self.label_7.setText(_translate("GenerateROIMaskTabContent", "/ "))
        self.maxSliceLabel.setText(_translate("GenerateROIMaskTabContent", "x"))
        self.label_11.setText(_translate("GenerateROIMaskTabContent", "Select dimension:"))
        self.label_14.setText(_translate("GenerateROIMaskTabContent", "(Default is <mask_name>_<dim>_<slice>.nii.gz)"))
        self.selectOutputFileInput.setText(_translate("GenerateROIMaskTabContent", "Browse"))
        self.label_12.setText(_translate("GenerateROIMaskTabContent", "Select slice:"))
        self.generateButton.setText(_translate("GenerateROIMaskTabContent", "Generate"))
        self.viewButton.setText(_translate("GenerateROIMaskTabContent", "View ROI"))




"""Definitions of the static data parameters.

These parameters are in usage similar to fixed free parameters. They are defined as static data parameters to
make clear that they are meant to carry additional observational data about a problem.

Please choose the parameter type for a model and parameter carefully since the type signifies how the parameter and
its data are handled during model construction.

"""
from mdt.models.parameters import StaticMapParameterConfig

__author__ = 'Robbert Harms'
__date__ = "2016-02-14"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class b1_static(StaticMapParameterConfig):

    data_type = 'mot_float_type'
    value = 1


class fa_static(StaticMapParameterConfig):

    data_type = 'mot_float_type'
    value = 1


class Sw_static(StaticMapParameterConfig):
    """This parameter is created only for linear T1 decay fitting of GRE data with variable flip angle.

    S_weighted is defined as the input data divided by the
        :math:`tan(flip_angle) ->  S_weighted = data / tan (flip_angle * B1_map)`
    """

    data_type = 'mot_float_type'
    value = 1


class T1_static(StaticMapParameterConfig):

    data_type = 'mot_float_type'


class T2_static(StaticMapParameterConfig):

    data_type = 'mot_float_type'


class T2s_static(StaticMapParameterConfig):

    data_type = 'mot_float_type'



from mdt.models.compartments import CompartmentConfig

__author__ = 'Robbert Harms'
__date__ = "2015-06-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ExpT1DecTR(CompartmentConfig):

    parameter_list = ('TR', 'T1')
    cl_code = 'return abs(1 - exp(-TR / T1));'



"""Definitions of the free parameters.

The free parameters are meant to be used for parameters that one wants to optimize. They can be fixed to a certain
value to disable them from being optimized in a given situation, but they remain classified as 'optimizable' parameters.

Please choose the parameter type for a model and parameter carefully since the type signifies how the parameter and
its data are handled during model construction.
"""

import numpy as np
from mdt.models.parameters import FreeParameterConfig
from mot.model_building.parameter_functions.priors import AbsSinHalfPrior, AbsSinPrior
from mot.model_building.parameter_functions.proposals import GaussianProposal, CircularGaussianProposal
from mot.model_building.parameter_functions.sample_statistics import CircularGaussianPSS
from mot.model_building.parameter_functions.transformations import ClampTransform, AbsModPiTransform, \
    SinSqrClampTransform, CosSqrClampTransform

__author__ = 'Robbert Harms'
__date__ = "2015-12-12"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"



class s0(FreeParameterConfig):

    init_value = 1e4
    lower_bound = 1e-5
    upper_bound = 1e10
    parameter_transform = ClampTransform()
    sampling_proposal = GaussianProposal(std=25.0)


class T1(FreeParameterConfig):

    init_value = 0.03
    lower_bound = 0.0
    upper_bound = 4.0
    parameter_transform = ClampTransform()
    sampling_proposal = GaussianProposal(0.0001)


class E1(FreeParameterConfig):
    """This parameter is defined *only* for linear decay T1 fitting in GRE data *with* TR constant.

    This parameter is also defined in the SSFP equation. However, in SSFP this parameter is from the protocol (!)
        E1 = exp( -TR / T1 ).
    After estimation of this parameter, T1 can be recovered by applying the next equation:
        -TR / log( E1 ).
    """

    init_value = 0.37
    lower_bound = 0.0
    upper_bound = 1.0
    parameter_transform = ClampTransform()
    sampling_proposal = GaussianProposal(0.0001)


class T2(FreeParameterConfig):

    init_value = 0.01
    lower_bound = 0.0
    upper_bound = 2.0
    parameter_transform = ClampTransform()
    sampling_proposal = GaussianProposal(0.0001)


class T2s(FreeParameterConfig):

    init_value = 0.01
    lower_bound = 0.0
    upper_bound = 1.0
    parameter_transform = ClampTransform()
    sampling_proposal = GaussianProposal(0.0001)


class R1(FreeParameterConfig):
    """R1 = 1/T1, for linear T1Dec or other models. """

    init_value = 2
    lower_bound = 0.25
    upper_bound = 100.0
    parameter_transform = ClampTransform()
    sampling_proposal = GaussianProposal(0.0001)


class R2(FreeParameterConfig):
    """R2 = 1/T2, for linear T2Dec or other models."""

    init_value = 5
    lower_bound = 0.5
    upper_bound = 500.0
    parameter_transform = ClampTransform()
    sampling_proposal = GaussianProposal(0.0001)


class R2s(FreeParameterConfig):
    """R2s = 1/T2s, for lineaR T2sDec or other models."""

    init_value = 10
    lower_bound = 1
    upper_bound = 50.0
    parameter_transform = ClampTransform()
    sampling_proposal = GaussianProposal(0.0001)


class theta(FreeParameterConfig):

    init_value = 1 / 2.0 * np.pi
    lower_bound = 0
    upper_bound = np.pi
    parameter_transform = AbsModPiTransform()
    sampling_proposal = CircularGaussianProposal(np.pi, 0.02)
    sampling_prior = AbsSinHalfPrior()
    sampling_statistics = CircularGaussianPSS()


class phi(FreeParameterConfig):

    init_value = 1 / 2.0 * np.pi
    lower_bound = 0
    upper_bound = np.pi
    parameter_transform = AbsModPiTransform()
    sampling_proposal = CircularGaussianProposal(np.pi, 0.02)
    sampling_prior = AbsSinPrior()
    sampling_statistics = CircularGaussianPSS()


class psi(FreeParameterConfig):

    init_value = 1 / 2.0 * np.pi
    lower_bound = 0
    upper_bound = np.pi
    parameter_transform = AbsModPiTransform()
    sampling_proposal = CircularGaussianProposal(np.pi, 0.02)
    sampling_prior = AbsSinPrior()
    sampling_statistics = CircularGaussianPSS()


class d(FreeParameterConfig):

    init_value = 1.7e-9
    lower_bound = 0
    upper_bound = 1.0e-8
    parameter_transform = SinSqrClampTransform()
    sampling_proposal = GaussianProposal(1e-14)


class dperp0(FreeParameterConfig):

    init_value = 1.7e-10
    lower_bound = 0
    upper_bound = 1.0e-8
    parameter_transform = SinSqrClampTransform()
    sampling_proposal = GaussianProposal(1e-15)


class dperp1(FreeParameterConfig):

    init_value = 1.7e-11
    lower_bound = 0
    upper_bound = 1.0e-8
    parameter_transform = SinSqrClampTransform()
    sampling_proposal = GaussianProposal(1e-15)


class R(FreeParameterConfig):

    init_value = 2.0e-6
    lower_bound = 1e-6
    upper_bound = 20e-6
    parameter_transform = CosSqrClampTransform()
    sampling_proposal = GaussianProposal(1e-6)


class kappa(FreeParameterConfig):

    init_value = 1
    lower_bound = 1e-5
    upper_bound = 2 * np.pi
    parameter_transform = CosSqrClampTransform()
    sampling_proposal = GaussianProposal(0.01)


# for use in the GDRCylinder model
class gamma_k(FreeParameterConfig):

    init_value = 1
    lower_bound = 0
    upper_bound = 20
    parameter_transform = SinSqrClampTransform()
    sampling_proposal = GaussianProposal(1.0)


# for use in the GDRCylinder model
class gamma_beta(FreeParameterConfig):

    init_value = 1
    lower_bound = 1.0e-7
    upper_bound = 3.0e-7
    parameter_transform = SinSqrClampTransform()
    sampling_proposal = GaussianProposal(1e-7)


# for use in the GDRCylinder model
class gamma_nmr_cyl(FreeParameterConfig):

    init_value = 5
    lower_bound = 1
    upper_bound = 10
    parameter_transform = SinSqrClampTransform()
    sampling_proposal = GaussianProposal(1)



from pkg_resources import resource_filename
from mot.cl_data_type import CLDataType
from mot.model_building.cl_functions.base import LibraryFunction
from mot.model_building.cl_functions.parameters import LibraryParameter

__author__ = 'Robbert Harms'
__date__ = "2015-06-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class NeumannCylPerpPGSESum(LibraryFunction):

    def __init__(self):
        super(NeumannCylPerpPGSESum, self).__init__(
            'double',
            'NeumannCylPerpPGSE',
            (LibraryParameter(CLDataType.from_string('mot_float_type'), 'Delta'),
             LibraryParameter(CLDataType.from_string('mot_float_type'), 'delta'),
             LibraryParameter(CLDataType.from_string('mot_float_type'), 'd'),
             LibraryParameter(CLDataType.from_string('mot_float_type'), 'R')),
            resource_filename(__name__, 'NeumannCylPerpPGSESum.h'),
            resource_filename(__name__, 'NeumannCylPerpPGSESum.cl'),
            {},
            ())



import glob
import os

import mdt
from mdt.batch_utils import SimpleBatchProfile, BatchFitProtocolLoader, SimpleSubjectInfo

__author__ = 'Robbert Harms'
__date__ = "2015-07-13"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"

meta_info = {'title': 'Directory per subject',
             'description': 'General layout for batch fitting with a folder per subject.',
             'directory_layout':
'''
Every subject has its own directory. For every type of file (protocol, bvec, bval, TE, Delta, delta, DWI, mask)
we use the first one found.

You can provide the noise standard deviation to use using a noise_std file containing a single float.

Example directory layout:
    /a/*.nii(.gz)
    /a/*bval*
    /a/*bvec*

    /b/...
    /c/...

Optional items:
    /*/prtcl
    /*/*.prtcl

    /*/TE (in seconds)
    /*/delta (in seconds)
    /*/Delta (in seconds)

    /*/*_mask.nii(.gz)

    /*/noise_std.{txt,nii,nii.gz}

The optional items TE, Delta and delta provide extra information about the protocol.
They should either contain exactly 1 value (for all protocol lines), or a value per protocol line.
'''}


class DirPerSubject(SimpleBatchProfile):

    def __init__(self):
        super(DirPerSubject, self).__init__()
        self.use_gradient_deviations = False

    def _get_subjects(self):
        subjects = []
        for subject_id in sorted([os.path.basename(f) for f in glob.glob(os.path.join(self._root_dir, '*'))]):
            pjoin = mdt.make_path_joiner(self._root_dir, subject_id)
            subject_info = self._get_subject_in_directory(subject_id, pjoin)
            if subject_info:
                subjects.append(subject_info)
        return subjects

    def _get_subject_in_directory(self, subject_id, pjoin):
        """Get the information about the given subject in the given directory.

        Args:
            subject_id (str): the id of this subject we are loading
            pjoin (PathJoiner): the path joiner pointing to the directory of that subject

        Returns:
            SimpleSubjectInfo or None: the subject info for this particular subject
        """
        niftis = glob.glob(pjoin('*.nii*'))
        dwis = list(filter(lambda v: all(name not in v for name in ['_mask', 'grad_dev', 'noise_std']), niftis))
        masks = list(filter(lambda v: '_mask' in v, niftis))
        grad_devs = list(filter(lambda v: 'grad_dev' in v, niftis))
        protocols = glob.glob(pjoin('*prtcl'))
        bvals = glob.glob(pjoin('*bval*'))
        bvecs = glob.glob(pjoin('*bvec*'))
        noise_std = self._autoload_noise_std(subject_id)

        if dwis:
            dwi_fname = dwis[0]

            mask_fnames = [pjoin('mask.nii')]
            mask_fnames.extend(masks)
            mask_fname = self._get_first_existing_nifti(mask_fnames)

            grad_dev = grad_devs[0] if grad_devs else None

            protocol_fname = protocols[0] if protocols else None
            bval_fname = bvals[0] if bvals else None
            bvec_fname = bvecs[0] if bvecs else None

            if dwi_fname and (protocol_fname or (bval_fname and bvec_fname)):
                protocol_loader = BatchFitProtocolLoader(
                    os.path.join(self._root_dir, subject_id),
                    protocol_fname=protocol_fname, bvec_fname=bvec_fname,
                    bval_fname=bval_fname)

                output_dir = self._get_subject_output_dir(subject_id, mask_fname)

                return SimpleSubjectInfo(subject_id, dwi_fname, protocol_loader, mask_fname, output_dir,
                                                  gradient_deviations=grad_dev,
                                                  use_gradient_deviations=self.use_gradient_deviations,
                                                  noise_std=noise_std)
        return None

    def __str__(self):
        return meta_info['title']



import os
import six
from mdt.protocols import Protocol, load_protocol, auto_load_protocol

__author__ = 'Robbert Harms'
__date__ = "2015-08-25"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


def autodetect_protocol_loader(data_source):
    """A function to get a protocol loader using the given data source.

    This tries to do auto detecting for the following data sources:

        - :class:`ProtocolLoader`
        - strings (filename or directory)
        - functions
        - protocol objects

    If a directory is given we try to auto use the protocol from sources in that directory.

    Args:
        data_source: the data source from which to get a protocol loader

    Returns:
        ProtocolLoader: a protocol loader instance.
    """
    if isinstance(data_source, ProtocolLoader):
        return data_source
    elif isinstance(data_source, six.string_types):
        if os.path.isfile(data_source):
            return ProtocolFromFileLoader(data_source)
        else:
            return ProtocolFromDirLoader(data_source)
    elif hasattr(data_source, '__call__'):
        return ProtocolFromFunctionLoader(data_source)
    elif isinstance(data_source, Protocol):
        return ProtocolDirectLoader(data_source)
    raise ValueError('The given data source could not be recognized.')


class ProtocolLoader(object):
    """Interface for loading protocols from different sources."""

    def get_protocol(self):
        """The public method used to get an instance of a protocol.

        Returns:
            Protocol: a protocol object
        """


class ProtocolFromFileLoader(ProtocolLoader):

    def __init__(self, filename):
        """Loads a protocol from the given filename.

        Args:
            filename (str): the filename to use the protocol from.
        """
        super(ProtocolFromFileLoader, self).__init__()
        self._filename = filename
        self._protocol = None

    def get_protocol(self):
        if self._protocol is None:
            self._protocol = load_protocol(self._filename)
        return self._protocol


class ProtocolFromDirLoader(ProtocolLoader):

    def __init__(self, directory):
        """Loads a protocol from the given filename.

        Args:
            directory (str): the directory to use the protocol from.
        """
        super(ProtocolFromDirLoader, self).__init__()
        self._directory = directory
        self._protocol = None

    def get_protocol(self):
        if self._protocol is None:
            self._protocol = auto_load_protocol(self._directory)
        return self._protocol


class ProtocolDirectLoader(ProtocolLoader):

    def __init__(self, protocol):
        """Adapter for returning an already loaded protocol.

        Args:
            protocol (Protocol): the loaded protocol to return.
        """
        super(ProtocolDirectLoader, self).__init__()
        self._protocol = protocol

    def get_protocol(self):
        return self._protocol


class ProtocolFromFunctionLoader(ProtocolLoader):

    def __init__(self, func):
        """Load a protocol from a callback function.

        This class may apply caching.

        Args:
            func: the callback function to call on the moment the protocol is to be loaded.
        """
        super(ProtocolFromFunctionLoader, self).__init__()
        self._func = func
        self._protocol = None

    def get_protocol(self):
        if self._protocol is None:
            self._protocol = self._func()
        return self._protocol






import six
import numpy as np
import numbers
from mdt.nifti import load_nifti

__author__ = 'Robbert Harms'
__date__ = "2015-08-25"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


def autodetect_noise_std_loader(data_source):
    """A function to get a noise std using the given data source.

    This tries to do auto detecting for the following data sources:

        - None: return 1
        - double: uses the given single value for all voxels
        - ndarray: use a value per voxel (this should not be a roi list, it should be an actual volume
            of the same size as the dataset)
        - string (and not 'auto'): a filename we will try to parse as a noise std
        - the string 'auto': try to estimate the noise std

    Args:
        data_source: the data source from which to get a noise std

    Returns:
        NoiseStdLoader: a noise std loader instance.
    """
    if isinstance(data_source, NoiseStdLoader):
        return data_source

    elif data_source is None:
        return NoiseEstimationLoader()

    elif isinstance(data_source, numbers.Number):
        return SingleValueNoiseStd(data_source)

    elif isinstance(data_source, np.ndarray):
        return VoxelWiseNoiseStd(data_source)

    elif isinstance(data_source, six.string_types):
        return LoadNoiseFromFile(data_source)

    raise ValueError('The given data source could not be recognized.')


class NoiseStdLoader(object):
    """Interface for loading a noise std from different sources."""

    def get_noise_std(self, problem_data):
        """The public method for getting the noise std from this loader.

        Args:
            problem_data (:class:`~mdt.utils.DMRIProblemData`): the dmri problem data in use.
                Some loaders might need this for loading the noise std.

        Returns:
            noise std, either a single value or an ndarray with a value per voxel
        """


class NoiseEstimationLoader(NoiseStdLoader):

    def __init__(self):
        """A loader that estimates the noise std from the problem data"""

    def get_noise_std(self, problem_data):
        from mdt.utils import estimate_noise_std
        return estimate_noise_std(problem_data)


class SingleValueNoiseStd(NoiseStdLoader):

    def __init__(self, noise_std):
        """Returns the given noise std"""
        self._noise_std = noise_std

    def get_noise_std(self, problem_data):
        return self._noise_std


class VoxelWiseNoiseStd(NoiseStdLoader):

    def __init__(self, noise_std_map):
        """Returns a noise std map with one value per voxel."""
        self._noise_std_map = noise_std_map

    def get_noise_std(self, problem_data):
        return self._noise_std_map


class LoadNoiseFromFile(NoiseStdLoader):

    def __init__(self, file_name):
        """Load a noise std from a file.

        This will try to detect if the given file is a text file with a single noise std, or if it is a nifti / map file
        with a voxel wise noise std.
        """
        self._file_name = file_name

    def get_noise_std(self, problem_data):
        if self._file_name[-4:] == '.txt':
            with open(self._file_name, 'r') as f:
                return float(f.read())

        return load_nifti(self._file_name).get_data()



from mdt.models.composite import DMRICompositeModelConfig

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ActiveAx(DMRICompositeModelConfig):

    ex_vivo_suitable = False
    description = 'The standard ActiveAx model'
    model_expression = '''
        S0 * ((Weight(w_ic) * CylinderGPD) +
              (Weight(w_ec) * Zeppelin) +
              (Weight(w_csf) * Ball))
    '''
    fixes = {'CylinderGPD.d': 1.7e-9,
             'Zeppelin.d': 1.7e-9,
             'Ball.d': 3.0e-9}
    dependencies = [('Zeppelin.dperp0', 'Zeppelin.d * (w_ec.w / (w_ec.w + w_ic.w))')]



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
"""Launches the MDT maps visualizer."""
import argparse
import os
import textwrap
from argcomplete.completers import FilesCompleter
from mdt.utils import init_user_settings
from mdt import view_maps
from mdt.visualization.maps.base import DataInfo
from mdt.shell_utils import BasicShellApplication, get_citation_message

__author__ = 'Robbert Harms'
__date__ = "2015-08-18"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class GUI(BasicShellApplication):

    def __init__(self):
        init_user_settings(pass_if_exists=True)

    def _get_arg_parser(self):
        description = textwrap.dedent(__doc__)
        description += get_citation_message()

        parser = argparse.ArgumentParser(description=description, formatter_class=argparse.RawTextHelpFormatter)
        parser.add_argument('dir', metavar='dir', type=str, nargs='?', help='the directory to use',
                            default=None).completer = FilesCompleter()

        parser.add_argument('-c', '--config', type=str,
                            help='Use the given initial configuration').completer = \
            FilesCompleter(['conf'], directories=False)

        parser.add_argument('-m', '--maximize', action='store_true', help="Maximize the shown window")
        parser.add_argument('--to-file', type=str, help="If set export the figure to the given filename")

        return parser

    def run(self, args):
        if args.dir:
            data = DataInfo.from_dir(os.path.realpath(args.dir))
        else:
            data = DataInfo.from_dir(os.getcwd())

        to_file = None
        if args.to_file:
            to_file = os.path.realpath(args.to_file)

        config = None
        if args.config:
            filename = os.path.realpath(args.config)
            if os.path.exists(filename):
                with open(filename, 'r') as f:
                    config = f.read()

        view_maps(data, config, show_maximized=args.maximize, to_file=to_file)

if __name__ == '__main__':
    GUI().start()



import copy
import os

from PyQt5.QtCore import pyqtSlot, Qt
from PyQt5.QtWidgets import QWidget, QAbstractItemView

from mdt.gui.maps_visualizer.actions import SetDimension, SetSliceIndex, SetVolumeIndex, SetColormap, SetRotate, \
    SetZoom, SetShowAxis, SetColorBarNmrTicks, SetMapsToShow, SetFont, SetInterpolation, SetFlipud, SetPlotTitle, \
    SetGeneralMask
from mdt.gui.maps_visualizer.design.ui_TabGeneral import Ui_TabGeneral
from mdt.gui.utils import blocked_signals, TimedUpdate
from mdt.visualization.maps.base import Zoom, Point, DataInfo, Font, MapPlotConfig

__author__ = 'Robbert Harms'
__date__ = "2016-09-03"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class TabGeneral(QWidget, Ui_TabGeneral):

    def __init__(self, controller, parent=None):
        super(TabGeneral, self).__init__(parent)
        self.setupUi(self)

        self._controller = controller
        self._controller.new_data.connect(self.set_new_data)
        self._controller.new_config.connect(self.set_new_config)

        self.general_display_order.setDragDropMode(QAbstractItemView.InternalMove)
        self.general_display_order.setSelectionMode(QAbstractItemView.SingleSelection)

        self.general_colormap.addItems(self._controller.get_config().get_available_colormaps())
        self.general_rotate.addItems(['0', '90', '180', '270'])
        self.general_rotate.setCurrentText(str(self._controller.get_config().rotate))

        self.general_DisplayOrder.set_collapse(True)
        self.general_Miscellaneous.set_collapse(True)
        self.general_Zoom.set_collapse(True)
        self.general_Font.set_collapse(True)

        self.general_dimension.valueChanged.connect(lambda v: self._controller.apply_action(SetDimension(v)))
        self.general_slice_index.valueChanged.connect(lambda v: self._controller.apply_action(SetSliceIndex(v)))
        self.general_volume_index.valueChanged.connect(lambda v: self._controller.apply_action(SetVolumeIndex(v)))
        self.general_colormap.currentIndexChanged.connect(
            lambda i: self._controller.apply_action(SetColormap(self.general_colormap.itemText(i))))
        self.general_rotate.currentIndexChanged.connect(
            lambda i: self._controller.apply_action(SetRotate(int(self.general_rotate.itemText(i)))))

        self._map_selection_timer = TimedUpdate(self._update_maps_to_show)
        self.general_map_selection.itemSelectionChanged.connect(
            lambda: self._map_selection_timer.add_delayed_callback(500))

        self.general_deselect_all_maps.clicked.connect(self._deleselect_all_maps)
        self.general_invert_map_selection.clicked.connect(self._invert_map_selection)

        self.general_zoom_x_0.valueChanged.connect(self._update_zoom)
        self.general_zoom_x_1.valueChanged.connect(self._update_zoom)
        self.general_zoom_y_0.valueChanged.connect(self._update_zoom)
        self.general_zoom_y_1.valueChanged.connect(self._update_zoom)

        self.plot_title.textEdited.connect(lambda txt: self._controller.apply_action(SetPlotTitle(txt)))

        self.general_zoom_reset.clicked.connect(lambda: self._controller.apply_action(SetZoom(Zoom.no_zoom())))
        self.general_zoom_fit.clicked.connect(self._zoom_fit)

        self.general_display_order.items_reordered.connect(self._reorder_maps)
        self.general_show_axis.clicked.connect(lambda: self._controller.apply_action(
            SetShowAxis(self.general_show_axis.isChecked())))
        self.general_colorbar_nmr_ticks.valueChanged.connect(
            lambda v: self._controller.apply_action(SetColorBarNmrTicks(v)))

        self.general_font_family.addItems(Font.font_names())
        self.general_font_family.currentTextChanged.connect(
            lambda v: self._controller.apply_action(SetFont(self._controller.get_config().font.get_updated(family=v))))

        self.general_font_size.valueChanged.connect(
            lambda: self._controller.apply_action(
                SetFont(self._controller.get_config().font.get_updated(size=self.general_font_size.value()))))

        self.general_interpolation.addItems(self._controller.get_config().get_available_interpolations())
        self.general_interpolation.currentTextChanged.connect(
            lambda v: self._controller.apply_action(SetInterpolation(v)))

        self.general_flipud.clicked.connect(lambda: self._controller.apply_action(
            SetFlipud(self.general_flipud.isChecked())))

        self.mask_name.currentIndexChanged.connect(self._update_mask_name)

    @pyqtSlot(DataInfo)
    def set_new_data(self, data_info):
        if data_info.directory:
            self.general_info_directory.setText(self._split_long_path_elements(data_info.directory))
        else:
            self.general_info_directory.setText('-')

        if len(data_info.maps):
            self.general_info_nmr_maps.setText(str(len(data_info.maps)))
        else:
            self.general_info_nmr_maps.setText('0')

        with blocked_signals(self.general_map_selection):
            self.general_map_selection.clear()
            self.general_map_selection.addItems(data_info.sorted_keys)
            for index, map_name in enumerate(data_info.sorted_keys):
                item = self.general_map_selection.item(index)
                item.setData(Qt.UserRole, map_name)

        with blocked_signals(self.mask_name):
            self.mask_name.clear()
            self.mask_name.insertItem(0, '-- None --')
            self.mask_name.insertItems(1, data_info.sorted_keys)

    @pyqtSlot(MapPlotConfig)
    def set_new_config(self, config):
        data_info = self._controller.get_data()
        map_names = config.maps_to_show

        with blocked_signals(self.general_dimension):
            try:
                max_dimension = data_info.get_max_dimension(map_names)
                self.general_dimension.setMaximum(max_dimension)
                self.maximumDimension.setText(str(max_dimension))
            except ValueError:
                self.general_dimension.setMaximum(0)
                self.maximumDimension.setText(str(0))
            self.general_dimension.setValue(config.dimension)

        with blocked_signals(self.general_slice_index):
            try:
                max_slice = data_info.get_max_slice_index(config.dimension, map_names)
                self.general_slice_index.setMaximum(max_slice)
                self.maximumIndex.setText(str(max_slice))
            except ValueError:
                self.general_slice_index.setMaximum(0)
                self.maximumIndex.setText(str(0))
            self.general_slice_index.setValue(config.slice_index)

        with blocked_signals(self.general_volume_index):
            try:
                max_volume = data_info.get_max_volume_index(map_names)
                self.general_volume_index.setMaximum(max_volume)
                self.maximumVolume.setText(str(max_volume))
            except ValueError:
                self.general_volume_index.setMaximum(0)
                self.maximumVolume.setText(str(0))
            self.general_volume_index.setValue(config.volume_index)

        with blocked_signals(self.general_colormap):
            self.general_colormap.setCurrentText(config.colormap)

        with blocked_signals(self.general_rotate):
            self.general_rotate.setCurrentText(str(config.rotate))

        if self.general_map_selection.count():
            for map_name, map_config in config.map_plot_options.items():
                if map_config.title:
                    index = data_info.sorted_keys.index(map_name)
                    item = self.general_map_selection.item(index)
                    item.setData(Qt.DisplayRole, map_name + ' (' + map_config.title + ')')

            self.general_map_selection.blockSignals(True)
            for index, map_name in enumerate(data_info.sorted_keys):
                item = self.general_map_selection.item(index)
                if item:
                    item.setSelected(map_name in map_names)
            self.general_map_selection.blockSignals(False)

        try:
            max_x = data_info.get_max_x_index(config.dimension, config.rotate, map_names)
            max_y = data_info.get_max_y_index(config.dimension, config.rotate, map_names)

            with blocked_signals(self.general_zoom_x_0, self.general_zoom_x_1,
                                 self.general_zoom_y_0, self.general_zoom_y_1):
                self.general_zoom_x_0.setMaximum(max_x)
                self.general_zoom_x_0.setValue(config.zoom.p0.x)

                self.general_zoom_x_1.setMaximum(max_x)
                self.general_zoom_x_1.setMinimum(config.zoom.p0.x)
                self.general_zoom_x_1.setValue(config.zoom.p1.x)

                self.general_zoom_y_0.setMaximum(max_y)
                self.general_zoom_y_0.setValue(config.zoom.p0.y)

                self.general_zoom_y_1.setMaximum(max_y)
                self.general_zoom_y_1.setMinimum(config.zoom.p0.y)
                self.general_zoom_y_1.setValue(config.zoom.p1.y)

                if config.zoom.p0.x == 0 and config.zoom.p1.x == 0:
                    self.general_zoom_x_1.setValue(max_x)

                if config.zoom.p0.y == 0 and config.zoom.p1.y == 0:
                    self.general_zoom_y_1.setValue(max_y)
        except ValueError:
            pass

        with blocked_signals(self.plot_title):
            self.plot_title.setText(config.title)

        with blocked_signals(self.general_display_order):
            self.general_display_order.clear()
            self.general_display_order.addItems(map_names)

            for index, map_name in enumerate(map_names):
                item = self.general_display_order.item(index)
                item.setData(Qt.UserRole, map_name)

                if map_name in config.map_plot_options and config.map_plot_options[map_name].title:
                    title = config.map_plot_options[map_name].title
                    item.setData(Qt.DisplayRole, map_name + ' (' + title + ')')

        with blocked_signals(self.general_show_axis):
            self.general_show_axis.setChecked(config.show_axis)

        with blocked_signals(self.general_colorbar_nmr_ticks):
            self.general_colorbar_nmr_ticks.setValue(config.colorbar_nmr_ticks)

        with blocked_signals(self.general_font_family):
            self.general_font_family.setCurrentText(config.font.family)

        with blocked_signals(self.general_font_size):
            self.general_font_size.setValue(config.font.size)

        with blocked_signals(self.general_interpolation):
            self.general_interpolation.setCurrentText(config.interpolation)

        with blocked_signals(self.general_flipud):
            self.general_flipud.setChecked(config.flipud)

        with blocked_signals(self.mask_name):
            if config.mask_name and config.mask_name in data_info.maps:
                for ind in range(self.mask_name.count()):
                    if self.mask_name.itemText(ind) == config.mask_name:
                        self.mask_name.setCurrentIndex(ind)
                        break
            else:
                self.mask_name.setCurrentIndex(0)

    @pyqtSlot()
    def _reorder_maps(self):
        items = [self.general_display_order.item(ind) for ind in range(self.general_display_order.count())]
        map_names = [item.data(Qt.UserRole) for item in items]
        self._controller.apply_action(SetMapsToShow(map_names))

    @pyqtSlot()
    def _update_maps_to_show(self):
        map_names = copy.copy(self._controller.get_config().maps_to_show)

        for item in [self.general_map_selection.item(ind) for ind in range(self.general_map_selection.count())]:
            map_name = item.data(Qt.UserRole)

            if item.isSelected():
                if map_name not in map_names:
                    self._insert_alphabetically(map_name, map_names)
            else:
                if map_name in map_names:
                    map_names.remove(map_name)

        self._controller.apply_action(SetMapsToShow(map_names))

    @pyqtSlot()
    def _deleselect_all_maps(self):
        self._controller.apply_action(SetMapsToShow([]))

    @pyqtSlot()
    def _invert_map_selection(self):
        self._controller.apply_action(SetMapsToShow(
            set(self._controller.get_data().maps.keys()).difference(set(self._controller.get_config().maps_to_show))))

    @pyqtSlot()
    def _zoom_fit(self):
        data_info = self._controller.get_data()
        config = self._controller.get_config()

        def add_padding(bounding_box, max_x, max_y):
            bounding_box[0].x = max(bounding_box[0].x - 1, 0)
            bounding_box[0].y = max(bounding_box[0].y - 1, 0)

            bounding_box[1].y = min(bounding_box[1].y + 2, max_y)
            bounding_box[1].x = min(bounding_box[1].x + 2, max_x)

            return bounding_box

        if config.maps_to_show or len(data_info.maps):
            bounding_box = data_info.get_bounding_box(config.dimension, config.slice_index,
                                                      config.volume_index, config.rotate, config.maps_to_show)

            max_y = data_info.get_max_y_index(config.dimension, rotate=config.rotate, map_names=config.maps_to_show)
            max_x = data_info.get_max_x_index(config.dimension, rotate=config.rotate, map_names=config.maps_to_show)

            if not config.flipud:
                # Since the renderer plots with a left top coordinate system,
                # we need to flip the y coordinates upside down by default.
                tmp = max_y - bounding_box[0].y
                bounding_box[0].y = max_y - bounding_box[1].y
                bounding_box[1].y = tmp

            bounding_box = add_padding(bounding_box, max_x, max_y)

            self._controller.apply_action(SetZoom(Zoom(*bounding_box)))

    @pyqtSlot()
    def _update_zoom(self):
        np0x, np0y = self.general_zoom_x_0.value(), self.general_zoom_y_0.value()
        np1x, np1y = self.general_zoom_x_1.value(), self.general_zoom_y_1.value()

        if np0x > np1x:
            np1x = np0x
        if np0y > np1y:
            np1y = np0y

        self._controller.apply_action(SetZoom(Zoom.from_coords(np0x, np0y, np1x, np1y)))

    @staticmethod
    def _insert_alphabetically(new_item, item_list):
        for ind, item in enumerate(item_list):
            if item > new_item:
                item_list.insert(ind, new_item)
                return
        item_list.append(new_item)

    @pyqtSlot(int)
    def _update_mask_name(self, index):
        if index == 0:
            self._controller.apply_action(SetGeneralMask(None))
        else:
            self._controller.apply_action(SetGeneralMask(self.mask_name.itemText(index)))

    def _split_long_path_elements(self, original_path, max_single_element_length=25):
        """Split long path elements into smaller ones using spaces

        Args:
            original_path (str): the path you want to split
            max_single_element_length (int): the maximum length allowed per path component (folders and filename).

        Returns:
            str: the same path but with spaces in long path elements. The result will no longer be a valid path.
        """

        def split(p):
            listing = []

            def _split(el):
                if el:
                    head, tail = os.path.split(el)
                    if not tail:
                        listing.append(head)
                    else:
                        _split(head)
                        listing.append(tail)

            _split(p)
            return listing

        elements = list(split(original_path))
        new_elements = []

        for el in elements:
            if len(el) > max_single_element_length:
                item = ''
                for i in range(0, len(el), max_single_element_length):
                    item += el[i:i + max_single_element_length] + ' '
                item = item[:-1]
                new_elements.append(item)
            else:
                new_elements.append(el)

        return os.path.join(*new_elements)



"""This module contains some functions that allow for generating simulated data.

The simulated data is on the level of diffusion MRI models, not on the level of simulated physical molecule interaction
as found in for example Camino.
"""
import numbers
import os
import numpy as np
import mdt
from mdt.nifti import write_nifti
from mdt.components_loader import NoiseSTDCalculatorsLoader
from mdt.utils import MockDMRIProblemData
from mot.cl_routines.mapping.calculate_model_estimates import CalculateModelEstimates

__author__ = 'Robbert Harms'
__date__ = "2016-03-17"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


def create_parameters_cube(primary_parameter_index, randomize_parameter_indices, grid_size,
                           default_values, lower_bounds, upper_bounds, dtype=np.float32, seed=None):
    """Create a simple 3d parameters cube.

    On the first dimension we put a linearly spaced primary parameter and on the second dimension we randomly change
    the other indicated parameters. The 3d dimension holds the parameter realizations for the other dimensions.

    Args:
        primary_parameter_index (int): the index of the primary parameter for the values on the first axis
        randomize_parameter_indices (list of int): the indices of the parameter we sample randomly using a uniform
            distribution on the half open interval between the [lower, upper) bounds. See np.random.uniform.
        grid_size (tuple of int): the size of the generated grid, the first value refers to the first dimension, the
            second to the second dimension.
        default_values (list of float): the default values for each of the parameters in the model
        lower_bounds (list of float): the lower bounds used for the generation of the grid
        upper_bounds (list of float): the upper bounds used for the generation of the grid
        dtype (dtype): the numpy data type for this grid
        seed (int): if given the seed for the random number generator, this makes the random parameters predictable.

    Returns:
        ndarray: a three dimensional cube for the parameters
    """
    grid = permutate_parameters([primary_parameter_index], default_values, lower_bounds, upper_bounds, grid_size[0],
                                dtype=dtype)
    grid = np.reshape(grid, (grid.shape[0], 1, grid.shape[1]))
    grid = np.repeat(grid, grid_size[1], axis=1)

    random_state = np.random.RandomState(seed)

    for param_ind in randomize_parameter_indices:
        grid[:, :, param_ind] = random_state.uniform(lower_bounds[param_ind], upper_bounds[param_ind], size=grid_size)

    return grid


def add_noise_realizations(signal_cube, nmr_noise_realizations, noise_sigma, seed=None):
    """Add noise realizations to a signal cube.

    The given signal cube should be a 3d matrix that contains on the last axis the signals per protocol line. All
    the other axis normally represent the variations of parameter values that generated the last dimension.

    This function inserts a new 3th dimension to that matrix, one to contain the variation over noise realisations. It
    will then make the signals rician distributed per index on that new 3th dimension.

    Args:
        signal_cube (ndarray): the 3d matrix with the signals
        nmr_noise_realizations (int): the number of noise realizations to use on the new third axis
        noise_sigma (float): the noise level, given by:
            noise_level = unweighted_signal_height / SNR
        seed (int): if given, the seed for the random number generation

    Returns:
        ndarray: a 4d cube with on the newly added third dimension the variating noise realizations
    """
    signals = np.reshape(signal_cube, signal_cube.shape[0: 2] + (1, signal_cube.shape[2]))
    signals = np.repeat(signals, nmr_noise_realizations, axis=2)
    return make_rician_distributed(signals, noise_sigma, seed=seed)


def simulate_signals_param_cube(model_name, protocol, parameters_cube):
    """Generate the signal for the given model for a generated parameters cube.

    Args:
        model_name (str): the name of the model we want to generate the values for
        protocol (Protocol): the protocol object we use for generating the signals
        parameters_cube (ndarray): the 3d matrix with the parameters for every problem instance

    Returns:
        signal estimates as a cube
    """
    parameters = np.reshape(parameters_cube, (-1, parameters_cube.shape[-1]))
    simulated_signals = simulate_signals(model_name, protocol, parameters)
    return np.reshape(simulated_signals, parameters_cube.shape[0:2] + (simulated_signals.shape[-1], ))


def permutate_parameters(var_params_ind, default_values, lower_bounds, upper_bounds, grid_size,
                         dtype=np.float32):
    """Generate the combination of parameters for a simulation.

    This is useful if you want to generate a list of different parameter combinations. You do not need to use
    this if you want to simulate only one parameter.

    This generates for each of the parameters of interest a linearly indexed range of parameter values
    starting with the lower bounds and ending at the upper bound (both inclusive). The length of the list is determined
    by the grid size per parameter. Next we create a matrix with the cartesian product of each of these parameters
    of interest and with all the other parameters set to their default value.

    Args:
        var_params_ind (list of int): the list of indices into the parameters. This indices the parameters
            we want to vary.
        default_values (list of float): the default values for each of the parameters in the model
        lower_bounds (list of float): the lower bounds used for the generation of the grid
        upper_bounds (list of float): the upper bounds used for the generation of the grid
        grid_size (int or list of int): the size of the grid. If a single int is given we assume a grid
            equal in all dimensions. If a list is given it should match the number of variable parameter indices
            and should contain a grid size for each parameter.
        dtype (dtype): the data type of the result matrix

    Returns:
        ndarray: the matrix with all combinations of the parameters of interest and with all other parameters set to
            the given default value.
    """
    if isinstance(grid_size, numbers.Number):
        grid_size = [int(grid_size)] * len(var_params_ind)

    result = np.reshape(default_values, [len(lower_bounds), 1]).astype(dtype)

    repeat_mult = 1
    for linear_ind, params_ind in enumerate(var_params_ind):
        result = np.tile(result, grid_size[linear_ind])
        result[params_ind] = np.repeat(np.linspace(lower_bounds[params_ind],
                                                   upper_bounds[params_ind],
                                                   grid_size[linear_ind]), repeat_mult)
        repeat_mult *= grid_size[linear_ind]

    return np.transpose(result)


def get_permuted_indices(nmr_var_params, grid_size):
    """Get for every parameter of interest the locations per parameter value.

    This is useful if you want to generate a list of different parameter combinations. You do not need to use
    this if you want to simulate only one parameter.

    Suppose you have three variable parameters and you generate all permutations using permutate_parameters(), then you
    might want to know for any given parameter and for any value of that parameter at which indices that parameter
    occurs. This function tells you where.

    Note, we could have taken the nmr_var_params from the grid size, but the grid size can be a single scalar for all
    params.

    Args:
        nmr_var_params (int): the number of variable parameters
        grid_size (int or list of int): the grid size for all or per parameter

    Returns:
        ndarray: per permutation the value index indexing the parameter value
    """
    indices = np.zeros((nmr_var_params, 1), dtype=np.int64)

    repeat_mult = 1
    for ind in range(nmr_var_params):
        indices = np.tile(indices, grid_size[ind])
        indices[ind, :] = np.repeat(np.arange(0, grid_size[ind]), repeat_mult)
        repeat_mult *= grid_size[ind]

    return np.transpose(indices)


def simulate_signals(model_name, protocol, parameters):
    """Generate the signal for the given model for each of the parameters.

    This function only accepts a 2d list of parameters. For a generated parameters cube use function
    simulate_signals_param_cube.

    Args:
        model_name (str): the name of the model we want to generate the values for
        protocol (Protocol): the protocol object we use for generating the signals
        parameters (ndarray): the 2d matrix with the parameters for every problem instance

    Returns:
        signal estimates
    """
    problem_data = MockDMRIProblemData(protocol, None, None, None)

    model = mdt.get_model(model_name)
    model.set_problem_data(problem_data)

    signal_evaluate = CalculateModelEstimates()
    return signal_evaluate.calculate(model, parameters)


def make_rician_distributed(signals, noise_level, seed=None):
    """Make the given signal Rician distributed.

    To calculate the noise level divide the signal of the unweighted volumes by the SNR you want. For example,
    for a unweighted signal b0=1e4 and a desired SNR of 20, you need an noise level of 1e4/20 = 500.

    Args:
        signals: the signals to make Rician distributed
        noise_level: the level of noise to add. The actual Rician stdev depends on the signal. See ricestat in
            the mathworks library. The noise level can be calculated using b0/SNR.
        seed (int): if given, the seed for the random number generation

    Returns:
        ndarray: Rician distributed signals.
    """
    random_state = np.random.RandomState(seed)
    x = noise_level * random_state.normal(size=signals.shape) + signals
    y = noise_level * random_state.normal(size=signals.shape)
    return np.sqrt(np.power(x, 2), np.power(y, 2)).astype(signals.dtype)


def list_2d_to_4d(item_list):
    """Convert a 2d signal/parameter list to a 4d volume.

    This appends two singleton volumes to the signal list to make it 4d.

    Args:
         item_list (2d ndarray): the list with on the first dimension every problem and on the second
            the signals per protocol line.

    Returns:
        ndarray: 4d ndarray of size (1, 1, n, p) where n is the number of problems and p the length of the protocol.
    """
    return np.reshape(item_list, (1, 1) + item_list.shape)


def save_data_volume(file_name, data):
    """Save the 3d/4d volume to the given file.

    Args:
        file_name (str): the output file name. If the directory does not exist we create one.
        data (ndarray): the 4d array to save.
    """
    if not os.path.isdir(os.path.dirname(file_name)):
        os.makedirs(os.path.dirname(file_name))
    write_nifti(data, None, file_name, affine=np.eye(4))


def save_2d_list_as_4d_volume(file_name, data):
    """Save the given 2d list with values as a 4d volume.

    This is a convenience function that calls list_2d_to_4d and volume4d_to_file after each other.

    Args:
        file_name (str): the output file name. If the directory does not exist we create one.
        data (ndarray): the 2d array to save
    """
    save_data_volume(file_name, list_2d_to_4d(data))


def get_unweighted_volumes(signals, protocol):
    """Get the signals and protocol for only the unweighted signals.

    Args:
        signals (ndarray): the matrix with for every problem (first dimension) the volumes (second dimension)
        protocol (Protocol): the protocol object

    Returns:
        tuple: unweighted signals and the protocol for only the unweighted indices.
    """
    unweighted_indices = protocol.get_unweighted_indices()

    unweighted_signals = signals[:, unweighted_indices]
    unweighted_protocol = protocol.get_new_protocol_with_indices(unweighted_indices)

    return unweighted_signals, unweighted_protocol


def estimate_noise_std(simulated_noisy_signals, protocol, noise_estimator_name='AllUnweightedVolumes'):
    """Estimate the noise on the noisy simulated dataset.

    This routine tries to estimate the noise level of the added noise. It first fits an S0 model to the data with
    a noise std of 1. It then removes this estimated S0 from the given signal and tries to estimate the noise std
    on the result.

    Args:
        simulated_noisy_signals (ndarray): the list with per problem the noisy simulated signal
        protocol (Protocol): the protocol object
        noise_estimator_name (str): the name of the noise estimator to use

    Returns:
        float: the noise standard deviation
    """
    mask = np.ones(simulated_noisy_signals.shape[0:3])

    loader = NoiseSTDCalculatorsLoader()
    cls = loader.get_class(noise_estimator_name)
    calculator = cls(mdt.load_problem_data(simulated_noisy_signals, protocol, mask))

    return calculator.estimate()



from mdt.models.compartments import CompartmentConfig

__author__ = 'Robbert Harms'
__date__ = "2015-06-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class AstroCylinders(CompartmentConfig):

    parameter_list = ('g', 'b', 'G', 'Delta', 'delta', 'd', 'R')
    dependency_list = ['MRIConstants',
                       'NeumannCylPerpPGSESum']
    cl_code = '''
        mot_float_type sum = NeumannCylPerpPGSESum(Delta, delta, d, R);

        mot_float_type lperp = (-2 * GAMMA_H_SQ * sum);
        mot_float_type lpar = -b * 1.0/pown(G, 2) * d;

        return (sqrt(M_PI) / (2 * G * sqrt(lperp - lpar)))
                    * exp(pown(G, 2) * lperp)
                    * erf(G * sqrt(lperp - lpar));
    '''



import matplotlib
import numpy as np

matplotlib.use('Qt5Agg')

from PyQt5.QtCore import QTimer

from mdt.visualization.maps.matplotlib_renderer import MapsVisualizer
from matplotlib.figure import Figure
from PyQt5 import QtWidgets, QtCore
from PyQt5.QtCore import pyqtSlot
from PyQt5.QtWidgets import QWidget, QVBoxLayout
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas

from mdt.gui.maps_visualizer.base import PlottingFrame
from mdt.visualization.maps.base import DataInfo, MapPlotConfig


class MatplotlibPlotting(PlottingFrame, QWidget):

    def __init__(self, controller, parent=None, plotting_info_viewer=None):
        super(MatplotlibPlotting, self).__init__(controller, plotting_info_viewer=plotting_info_viewer)

        self._controller.new_data.connect(self.set_new_data)
        self._controller.new_config.connect(self.set_new_config)

        self._auto_render = True

        self.figure = Figure()
        self.visualizer = MapsVisualizer(self._controller.get_data(), self.figure)
        self._axes_data = self.visualizer.render(self._controller.get_config())

        self.canvas = FigureCanvas(self.figure)
        self.canvas.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self.canvas.updateGeometry()

        layout = QVBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.canvas)
        self.setLayout(layout)

        self.setParent(parent)
        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.setFocus()

        self._redraw_timer = QTimer()
        self._redraw_timer.timeout.connect(self._timer_event)
        self._redraw_timer.timeout.connect(self._redraw_timer.stop)

        self._mouse_interaction = _MouseInteraction(self.figure, self._plotting_info_viewer)
        self._mouse_interaction.update_axes_data(self._axes_data)

        self._previous_config = None

        self.setMinimumWidth(100)

    def export_image(self, filename, width, height, dpi=100):
        width_inch = width / dpi
        height_inch = height / dpi

        figure = Figure(figsize=(width_inch, height_inch), dpi=dpi)
        visualizer = MapsVisualizer(self._controller.get_data(), figure)
        FigureCanvas(figure)

        visualizer.to_file(filename, self._controller.get_config(), dpi=dpi)

    def set_auto_rendering(self, auto_render):
        self._auto_render = auto_render

    def redraw(self):
        self._redraw()

    @pyqtSlot()
    def _timer_event(self):
        self._redraw()

    @pyqtSlot(DataInfo)
    def set_new_data(self, data_info):
        self.visualizer = MapsVisualizer(data_info, self.figure)
        self._redraw_timer.start(300)

    @pyqtSlot(MapPlotConfig)
    def set_new_config(self, configuration):
        if not self._previous_config or configuration.visible_changes(self._previous_config):
            self._previous_config = configuration
            if self._auto_render:
                self._redraw_timer.start(300)

    def _redraw(self):
        self.figure.clf()

        self._axes_data = self.visualizer.render(self._controller.get_config())
        self._mouse_interaction.update_axes_data(self._axes_data)

        self.figure.canvas.draw()


class _MouseInteraction(object):

    def __init__(self, figure, plotting_info_viewer):
        self.figure = figure
        self.plotting_info_viewer = plotting_info_viewer
        self._axes_data = []
        self.figure.canvas.mpl_connect('button_release_event', self._button_released)
        self.figure.canvas.mpl_connect('motion_notify_event', self._mouse_motion)

    def update_axes_data(self, axes_data):
        """Set the updated axes data. Needs to be called if the axes are updated.

        Args:
            axes_data (list of AxisData): the information about the axes
        """
        self._axes_data = axes_data

    def _button_released(self, event):
        axis_data = self._get_matching_axis_data(event.inaxes)
        if axis_data:
            x, y = int(np.round(event.xdata)), int(np.round(event.ydata))
            index = axis_data.coordinates_to_index(x, y)
            value = axis_data.get_value(index)
            # todo draw info box on the figure
            # print(x, y, index, value)

    def _mouse_motion(self, event):
        axis_data = self._get_matching_axis_data(event.inaxes)
        if axis_data:
            x, y = int(np.round(event.xdata)), int(np.round(event.ydata))
            index = axis_data.coordinates_to_index(x, y)
            value = axis_data.get_value(index)
            self.plotting_info_viewer.set_voxel_info((x, y), tuple(index), float(value))
        else:
            self.plotting_info_viewer.clear_voxel_info()

    def _get_matching_axis_data(self, axis):
        """Get the axis data matching the given axis.

        Args:
            Axis: the matplotlib axis to match

        Returns:
            AxisData: our data container for that axis
        """
        if axis:
            for axes_data in self._axes_data:
                if axes_data.axis == axis:
                    return axes_data
        return None



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'TabMapSpecific.ui'
#
# Created by: PyQt5 UI code generator 5.4.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_TabMapSpecific(object):
    def setupUi(self, TabMapSpecific):
        TabMapSpecific.setObjectName("TabMapSpecific")
        TabMapSpecific.resize(445, 534)
        self.gridLayout = QtWidgets.QGridLayout(TabMapSpecific)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setSpacing(0)
        self.gridLayout.setObjectName("gridLayout")
        self.scrollArea_2 = QtWidgets.QScrollArea(TabMapSpecific)
        self.scrollArea_2.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.scrollArea_2.setWidgetResizable(True)
        self.scrollArea_2.setObjectName("scrollArea_2")
        self.scrollAreaWidgetContents_2 = QtWidgets.QWidget()
        self.scrollAreaWidgetContents_2.setGeometry(QtCore.QRect(0, 0, 443, 532))
        self.scrollAreaWidgetContents_2.setObjectName("scrollAreaWidgetContents_2")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents_2)
        self.verticalLayout.setContentsMargins(6, 6, 6, 6)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setObjectName("verticalLayout")
        self.selectedMap = QtWidgets.QComboBox(self.scrollAreaWidgetContents_2)
        self.selectedMap.setObjectName("selectedMap")
        self.verticalLayout.addWidget(self.selectedMap)
        self.frame = QtWidgets.QFrame(self.scrollAreaWidgetContents_2)
        self.frame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame.setObjectName("frame")
        self.gridLayout_4 = QtWidgets.QGridLayout(self.frame)
        self.gridLayout_4.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_4.setSpacing(0)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.mapSpecificOptionsPosition = QtWidgets.QGridLayout()
        self.mapSpecificOptionsPosition.setSpacing(0)
        self.mapSpecificOptionsPosition.setObjectName("mapSpecificOptionsPosition")
        self.gridLayout_4.addLayout(self.mapSpecificOptionsPosition, 0, 0, 1, 1)
        self.verticalLayout.addWidget(self.frame)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.scrollArea_2.setWidget(self.scrollAreaWidgetContents_2)
        self.gridLayout.addWidget(self.scrollArea_2, 0, 0, 1, 1)

        self.retranslateUi(TabMapSpecific)
        QtCore.QMetaObject.connectSlotsByName(TabMapSpecific)
        TabMapSpecific.setTabOrder(self.scrollArea_2, self.selectedMap)

    def retranslateUi(self, TabMapSpecific):
        _translate = QtCore.QCoreApplication.translate
        TabMapSpecific.setWindowTitle(_translate("TabMapSpecific", "Form"))




from mdt.models.compartments import CompartmentConfig

__author__ = 'Francisco.Lagos'


class ExpT1ExpT2sGRE(CompartmentConfig):

    parameter_list = ('TR', 'TE', 'flip_angle', 'T1', 'T2s')
    cl_code = """
        return sin(flip_angle) * (1 - exp(-TR / T1)) / (1 - cos(flip_angle) * exp(-TR / T1)) * exp(-TE / T2s);
    """



import glob
import warnings
from copy import deepcopy

import nibabel
import numpy as np
import yaml
import matplotlib.font_manager
import mdt
import mdt.visualization.layouts
from mdt import nifti
from mdt.deferred_mappings import DeferredActionDict
from mdt.visualization.dict_conversion import StringConversion, \
    SimpleClassConversion, IntConversion, SimpleListConversion, BooleanConversion, \
    ConvertDictElements, ConvertDynamicFromModule, FloatConversion, WhiteListConversion
from mdt.visualization.layouts import Rectangular

__author__ = 'Robbert Harms'
__date__ = "2016-09-02"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class MapPlotConfig(object):

    def __init__(self, dimension=2, slice_index=0, volume_index=0, rotate=90, colormap='hot', maps_to_show=None,
                 font=None, grid_layout=None, colorbar_nmr_ticks=10, show_axis=True, zoom=None,
                 map_plot_options=None, interpolation='bilinear', flipud=None,
                 title=None, mask_name=None):
        """Container for all plot related settings.

        Args:
            dimension (int): the dimension we are viewing
            slice_index (int): the slice in the dimension we are viewing
            volume_index (int): in the case of multiple volumes (4th dimension) which index we are in.
            rotate (int): the rotation factor, multiple of 90. By default we rotate 90 degrees to
                show most in-vivo datasets in a natural way.
            colormap (str): the name of the colormap to use
            maps_to_show (list of str): the names of the maps to show
            font (int): the font settings
            grid_layout (GridLayout): the layout of the grid
            colorbar_nmr_ticks (int): the number of ticks on the colorbar
            show_axis (bool): if we show the axis or not
            zoom (Zoom): the zoom setting for all the plots
            map_plot_options (dict): per map the map specific plot options
            interpolation (str): one of the available interpolations
            flipud (boolean): if True we flip the image upside down
            title (str): the title to this plot
            mask_name (str): the name of the mask to apply to the maps prior to display
        """
        super(MapPlotConfig, self).__init__()
        self.dimension = dimension
        self.slice_index = slice_index
        self.volume_index = volume_index
        self.rotate = rotate
        self.colormap = colormap
        self.maps_to_show = maps_to_show or []
        self.zoom = zoom or Zoom.no_zoom()
        self.font = font or Font()
        self.colorbar_nmr_ticks = colorbar_nmr_ticks
        self.show_axis = show_axis
        if self.show_axis is None:
            self.show_axis = True
        self.grid_layout = grid_layout or Rectangular()
        self.interpolation = interpolation or 'bilinear'
        self.flipud = flipud
        if self.flipud is None:
            self.flipud = False
        self.map_plot_options = map_plot_options or {}
        self.title = title
        self.mask_name = mask_name

        if interpolation not in self.get_available_interpolations():
            raise ValueError('The given interpolation ({}) is not supported.'.format(interpolation))

        if self.colormap not in self.get_available_colormaps():
            raise ValueError('The given colormap ({}) is not supported.'.format(self.colormap))

        if self.rotate not in [0, 90, 180, 270]:
            raise ValueError('The given rotation ({}) is not supported, use 90 '
                             'degree angles within 360.'.format(self.rotate))

        if self.dimension is None:
            raise ValueError('The dimension can not be None.')

        if self.slice_index is None:
            raise ValueError('The slice index can not be None.')

        if self.volume_index is None:
            raise ValueError('The volume index can not be None.')

        if self.rotate is None:
            raise ValueError('The rotation can not be None.')

        if self.dimension < 0:
            raise ValueError('The dimension can not be smaller than 0, {} given.'.format(self.dimension))

    @classmethod
    def get_available_interpolations(cls):
        return get_available_interpolations()

    @classmethod
    def get_available_colormaps(cls):
        return get_available_colormaps()

    @classmethod
    def get_conversion_info(cls):
        return SimpleClassConversion(cls, cls._get_attribute_conversions())

    @classmethod
    def _get_attribute_conversions(cls):
        return {'dimension': IntConversion(),
                'slice_index': IntConversion(),
                'volume_index': IntConversion(),
                'rotate': IntConversion(),
                'colormap': StringConversion(),
                'maps_to_show': SimpleListConversion(),
                'zoom': Zoom.get_conversion_info(),
                'font': Font.get_conversion_info(),
                'colorbar_nmr_ticks': IntConversion(),
                'show_axis': BooleanConversion(),
                'map_plot_options': ConvertDictElements(SingleMapConfig.get_conversion_info()),
                'grid_layout': ConvertDynamicFromModule(mdt.visualization.layouts),
                'interpolation': WhiteListConversion(cls.get_available_interpolations(), 'bilinear'),
                'flipud': BooleanConversion(allow_null=False),
                'title': StringConversion(),
                'mask_name': StringConversion()
                }

    @classmethod
    def from_yaml(cls, text):
        return cls.get_conversion_info().from_dict(yaml.load(text))

    @classmethod
    def from_dict(cls, config_dict):
        return cls.get_conversion_info().from_dict(config_dict)

    def to_dict(self):
        return self.get_conversion_info().to_dict(self)

    def to_yaml(self):
        return yaml.safe_dump(self.get_conversion_info().to_dict(self))

    def visible_changes(self, old_config):
        """Checks if there are any visible changes between this configuration and the other.

        This method can implement knowledge that allows the visualization routine to check if it
        would need to update the plot or not.

        It expects that the configuration you wish to display is the one on which this method is called.

        Args:
            old_config (MapPlotConfig): the previous configuration

        Returns:
            bool: if the differences between this configuration and the other would result in visible differences.
        """
        def visible_difference_in_map_plot_options():
            for key in set(self.map_plot_options.keys()):
                if key in self.maps_to_show:
                    if key not in old_config.map_plot_options:
                        return True
                    if self.map_plot_options[key].visible_changes(old_config.map_plot_options[key]):
                        return True
            return False

        if any(getattr(self, key) != getattr(old_config, key) for key in
               filter(lambda key: key != 'map_plot_options', self.__dict__)):
            return True

        return visible_difference_in_map_plot_options()

    def validate(self, data_info):
        if data_info.maps:
            self._validate_maps_to_show(data_info)
            self._validate_dimension(data_info)

            for key in self.__dict__:
                if hasattr(self, '_validate_' + key):
                    getattr(self, '_validate_' + key)(data_info)
        return self

    def _validate_maps_to_show(self, data_info):
        if any(map(lambda k: k not in data_info.maps, self.maps_to_show)):
            raise ValueError('One or more of the given maps to show is not in the data: {}'.
                             format(set(self.maps_to_show).difference(set(data_info.maps))))

    def _validate_dimension(self, data_info):
        max_dim = data_info.get_max_dimension(map_names=self.maps_to_show)
        if self.dimension is None or self.dimension > max_dim:
            raise ValueError('The dimension ({}) can not be higher than {}.'.format(self.dimension, max_dim))

    def _validate_slice_index(self, data_info):
        max_slice_index = data_info.get_max_slice_index(self.dimension, map_names=self.maps_to_show)
        if self.slice_index is None or self.slice_index > max_slice_index or self.slice_index < 0:
            raise ValueError('The slice index ({}) can not be higher than '
                             '{} or lower than 0.'.format(self.slice_index, max_slice_index))

    def _validate_volume_index(self, data_info):
        max_volume_index = data_info.get_max_volume_index(map_names=self.maps_to_show)
        if self.volume_index > max_volume_index or self.volume_index < 0:
            raise ValueError('The volume index ({}) can not be higher than '
                             '{} or lower than 0.'.format(self.volume_index, max_volume_index))

    def _validate_zoom(self, data_info):
        max_x = data_info.get_max_x_index(self.dimension, self.rotate)
        max_y = data_info.get_max_y_index(self.dimension, self.rotate)

        if self.zoom.p1.x > max_x:
            raise ValueError('The zoom maximum x ({}) can not be larger than {}'.format(self.zoom.p1.x, max_x))

        if self.zoom.p1.y > max_y:
            raise ValueError('The zoom maximum y ({}) can not be larger than {}'.format(self.zoom.p1.y, max_y))

    def _validate_mask_name(self, data_info):
        if self.mask_name:
            if self.mask_name not in data_info.maps:
                raise ValueError('The given global mask is not found in the list of maps.')

    def _validate_map_plot_options(self, data_info):
        for key in self.map_plot_options:
            if key not in data_info.maps:
                del self.map_plot_options[key]

        for key, value in self.map_plot_options.items():
            if value is not None:
                self.map_plot_options[key] = value.validate(data_info)

    def __repr__(self):
        return str(self.get_conversion_info().to_dict(self))

    def __eq__(self, other):
        if not isinstance(other, MapPlotConfig):
            return NotImplemented
        for key, value in self.__dict__.items():
            if value != getattr(other, key):
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)


class SingleMapConfig(object):

    def __init__(self, title=None, scale=None, clipping=None, colormap=None, colorbar_label=None, title_spacing=None,
                 mask_name=None):
        """Creates the configuration for a single map plot.

        Args:
            title (str): the title of this plot, can contain latex using the matplotlib latex syntax
            scale (Scale): the scaling for the values in this map
            clipping (Clipping): the clipping to apply to the values prior to plotting
            colormap (str): the matplotlib colormap to use
            colorbar_label (str): the label for the colorbar
            title_spacing (float): the spacing between the top of the plots and the title
            mask_name (str): the name of the mask used to mask the data prior to visualization
        """
        super(SingleMapConfig, self).__init__()
        self.title = title
        self.title_spacing = title_spacing
        self.scale = scale or Scale()
        self.clipping = clipping or Clipping()
        self.colormap = colormap
        self.colorbar_label = colorbar_label
        self.mask_name = mask_name

        if self.colormap is not None and self.colormap not in self.get_available_colormaps():
            raise ValueError('The given colormap ({}) is not supported.'.format(self.colormap))

    @classmethod
    def get_conversion_info(cls):
        return SimpleClassConversion(cls, cls._get_attribute_conversions())

    @classmethod
    def _get_attribute_conversions(cls):
        return {'title': StringConversion(),
                'scale': Scale.get_conversion_info(),
                'clipping': Clipping.get_conversion_info(),
                'colormap': StringConversion(),
                'colorbar_label': StringConversion(),
                'title_spacing': FloatConversion(),
                'mask_name': StringConversion()}

    @classmethod
    def get_available_colormaps(cls):
        return get_available_colormaps()

    @classmethod
    def from_yaml(cls, text):
        return cls.get_conversion_info().from_dict(yaml.load(text))

    @classmethod
    def from_dict(cls, config_dict):
        return cls.get_conversion_info().from_dict(config_dict)

    def to_dict(self):
        return self.get_conversion_info().to_dict(self)

    def to_yaml(self):
        return yaml.safe_dump(self.get_conversion_info().to_dict(self))

    def visible_changes(self, old_config):
        """Checks if there are any visible changes between this configuration and the other.

        This method can implement knowledge that allows the visualization routine to check if it
        would need to update the plot or not.

        It expects that the configuration you wish to display is the one on which this method is called.

        Args:
            old_config (SingleMapConfig): the previous configuration

        Returns:
            bool: if the differences between this configuration and the other would result in visible differences.
        """
        def filtered_attributes():
            filtered = ['scale', 'clipping']
            return [key for key in self.__dict__ if key not in filtered]

        def visible_changes_in_scale():
            return self.scale.visible_changes(old_config.scale)

        def visible_changes_in_clipping():
            return self.clipping.visible_changes(old_config.clipping)

        if any(getattr(self, key) != getattr(old_config, key) for key in filtered_attributes()):
            return True

        return visible_changes_in_clipping() or visible_changes_in_scale()

    def validate(self, data_info):
        for key in self.__dict__:
            if hasattr(self, '_validate_' + key):
                getattr(self, '_validate_' + key)(data_info)
        return self

    def __repr__(self):
        return str(self.get_conversion_info().to_dict(self))

    def __eq__(self, other):
        if not isinstance(other, SingleMapConfig):
            return NotImplemented

        for key, value in self.__dict__.items():
            if value != getattr(other, key):
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)


class Zoom(object):

    def __init__(self, p0, p1):
        """Container for zooming a map between the two given points.

        Args:
            p0 (Point): the lower left corner of the zoomed area
            p1 (Point): the upper right corner of the zoomed area
        """
        self.p0 = p0
        self.p1 = p1

        if p0.x > p1.x or p0.y > p1.y:
            raise ValueError('The lower left point ({}, {}) should be smaller than the upper right point ({}, {})'.
                             format(p0.x, p0.y, p1.x, p1.y))

        if p0.x < 0 or p0.y < 0 or p1.x < 0 or p1.y < 0:
            raise ValueError('The zoom box ({}, {}), ({}, {}) can not '
                             'be negative in any way.'.format(p0.x, p0.y, p1.x, p1.y))

        if self.p0 is None or self.p1 is None:
            raise ValueError('One of the zoom points is None.')

    @classmethod
    def from_coords(cls, x0, y0, x1, y1):
        return cls(Point(x0, y0), Point(x1, y1))

    @classmethod
    def no_zoom(cls):
        return cls(Point(0, 0), Point(0, 0))

    @classmethod
    def get_conversion_info(cls):
        return SimpleClassConversion(cls, cls._get_attribute_conversions())

    @classmethod
    def _get_attribute_conversions(cls):
        point_converter = Point.get_conversion_info()
        return {'p0': point_converter,
                'p1': point_converter}

    def get_rotated(self, rotation, x_dimension, y_dimension):
        """Return a new Zoom instance rotated with the given factor.

        This rotates the zoom box in the same way as the image is rotated.

        Args:
            rotation (int): the rotation by which to rotate in steps of 90 degrees
            x_dimension (int): the dimension of the image in the x coordinate
            y_dimension (int): the dimension of the image in the y coordinate

        Returns:
            Zoom: the rotated instance
        """
        dimensions = [x_dimension, y_dimension]
        p0 = self.p0
        p1 = self.p1

        nmr_90_rotations = rotation % 360 // 90

        for _ in range(nmr_90_rotations):
            dimensions = np.roll(dimensions, 1)

            new_p0 = Point(np.min([dimensions[0] - p0.y, dimensions[0] - p1.y]), np.min([p0.x, p1.x]))
            new_p1 = Point(np.max([dimensions[0] - p0.y, dimensions[0] - p1.y]), np.max([p0.x, p1.x]))

            p0 = new_p0
            p1 = new_p1

        if p0.x >= dimensions[0] - 1 or p0.x < 0:
            p0 = p0.get_updated(x=0)

        if p0.y >= dimensions[1] - 1 or p0.y < 0:
            p0 = p0.get_updated(y=0)

        if p1.x >= dimensions[0] - 1:
            p1 = p1.get_updated(x=dimensions[0] - 1)

        if p1.y >= dimensions[1] - 1:
            p1 = p1.get_updated(y=dimensions[1] - 1)

        return Zoom(p0, p1)

    def apply(self, data):
        """Apply the zoom to the given 2d array and return the new array.

        Args:
            data (ndarray): the data to zoom in on
        """
        correct = self.p0.x < data.shape[1] and self.p1.x < data.shape[1] \
                  and self.p0.y < data.shape[0] and self.p1.y < data.shape[0] \
                  and self.p0.x < self.p1.x and self.p0.y < self.p1.y
        if correct:
            return data[self.p0.y:self.p1.y, self.p0.x:self.p1.x]
        return data

    def __repr__(self):
        return str(self.get_conversion_info().to_dict(self))

    def __eq__(self, other):
        if not isinstance(other, Zoom):
            return NotImplemented

        for key, value in self.__dict__.items():
            if value != getattr(other, key):
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)


class Point(object):

    def __init__(self, x, y):
        """Container for a single point"""
        self.x = x
        self.y = y

    def get_updated(self, **kwargs):
        """Get a new Point object with updated arguments.

        Args:
            **kwargs (dict): the new keyword values, when given these take precedence over the current ones.

        Returns:
            Point: a new scale with updated values.
        """
        new_values = dict(x=self.x, y=self.y)
        new_values.update(**kwargs)
        return Point(**new_values)

    @classmethod
    def get_conversion_info(cls):
        return SimpleClassConversion(cls, cls._get_attribute_conversions())

    @classmethod
    def _get_attribute_conversions(cls):
        return {'x': IntConversion(allow_null=False),
                'y': IntConversion(allow_null=False)}

    def rotate90(self, nmr_rotations):
        """Rotate this point around a 90 degree angle

        Args:
            nmr_rotations (int): the number of 90 degreee rotations, can be negative

        Returns:
            Point: the rotated point
        """

        def rotate_coordinate(x, y, nmr_rotations):
            rotation_matrix = np.array([[0, -1],
                                        [1, 0]])
            rx, ry = x, y
            for rotation in range(1, nmr_rotations + 1):
                rx, ry = rotation_matrix.dot([rx, ry])
            return rx, ry

        return Point(*rotate_coordinate(self.x, self.y, nmr_rotations))

    def __repr__(self):
        return 'Point(x={}, y={})'.format(self.x, self.y)

    def __eq__(self, other):
        if not isinstance(other, Point):
            return NotImplemented

        for key, value in self.__dict__.items():
            if value != getattr(other, key):
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)


class Clipping(object):

    def __init__(self, vmin=0, vmax=0, use_min=False, use_max=False):
        """Container for the map clipping information"""
        self.vmin = vmin
        self.vmax = vmax
        self.use_min = use_min
        self.use_max = use_max

        if use_min and use_max and vmin > vmax:
            raise ValueError('The minimum clipping ({}) can not be larger than the maximum clipping({})'.format(
                vmin, vmax))

    def apply(self, data):
        """Apply the clipping to the given 2d array and return the new array.

        Args:
           data (ndarray): the data to clip
        """
        if self.use_max or self.use_min:
            clipping_min = data.min()
            if self.use_min:
                clipping_min = self.vmin

            clipping_max = data.max()
            if self.use_max:
                clipping_max = self.vmax

            return np.clip(data, clipping_min, clipping_max)

        return data

    def visible_changes(self, old_clipping):
        """Checks if there are any visible changes between this clipping and the other.

        This method can implement knowledge that allows the visualization routine to check if it
        would need to update the plot or not.

        It expects that the clipping you wish to use is the one on which this method is called.

        Args:
            old_clipping (Clipping): the previous clipping

        Returns:
            bool: if the differences between this clipping and the other would result in visible differences.
        """
        if self.use_min and not old_clipping.use_min:
            return True
        if self.use_max and not old_clipping.use_max:
            return True

        def visible_changes_in_min():
            if self.vmin == old_clipping.vmin:
                return False
            else:
                return self.use_min

        def visible_changes_in_max():
            if self.vmax == old_clipping.vmax:
                return False
            else:
                return self.use_max

        return visible_changes_in_max() or visible_changes_in_min()

    def get_updated(self, **kwargs):
        """Get a new Clipping object with updated arguments.

        Args:
            **kwargs (dict): the new keyword values, when given these take precedence over the current ones.

        Returns:
            Clipping: a new scale with updated values.
        """
        new_values = dict(vmin=self.vmin, vmax=self.vmax, use_min=self.use_min, use_max=self.use_max)
        new_values.update(**kwargs)
        return Clipping(**new_values)

    @classmethod
    def get_conversion_info(cls):
        return SimpleClassConversion(cls, cls._get_attribute_conversions())

    @classmethod
    def _get_attribute_conversions(cls):
        return {'vmax': FloatConversion(allow_null=False),
                'vmin': FloatConversion(allow_null=False),
                'use_min': BooleanConversion(allow_null=False),
                'use_max': BooleanConversion(allow_null=False)}

    def __eq__(self, other):
        if not isinstance(other, Clipping):
            return NotImplemented

        for key, value in self.__dict__.items():
            if value != getattr(other, key):
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)


class Scale(object):

    def __init__(self, vmin=0, vmax=0, use_min=False, use_max=False):
        """Container the map scaling information"""
        self.vmin = vmin
        self.vmax = vmax
        self.use_min = use_min
        self.use_max = use_max

        if use_min and use_max and vmin > vmax:
            raise ValueError('The minimum scale ({}) can not be larger than the maximum scale ({})'.format(vmin, vmax))

    @classmethod
    def get_conversion_info(cls):
        return SimpleClassConversion(cls, cls._get_attribute_conversions())

    @classmethod
    def _get_attribute_conversions(cls):
        return {'vmax': FloatConversion(allow_null=False),
                'vmin': FloatConversion(allow_null=False),
                'use_min': BooleanConversion(allow_null=False),
                'use_max': BooleanConversion(allow_null=False)}

    def visible_changes(self, old_scale):
        """Checks if there are any visible changes between this scale and the other.

        This method can implement knowledge that allows the visualization routine to check if it
        would need to update the plot or not.

        It expects that the scale you wish to use is the one on which this method is called.

        Args:
            old_scale (Scale): the previous scale

        Returns:
            bool: if the differences between this scale and the other would result in visible differences.
        """
        if self.use_min and not old_scale.use_min:
            return True
        if self.use_max and not old_scale.use_max:
            return True

        def visible_changes_in_min():
            if self.vmin == old_scale.vmin:
                return False
            else:
                return self.use_min

        def visible_changes_in_max():
            if self.vmax == old_scale.vmax:
                return False
            else:
                return self.use_max

        return visible_changes_in_max() or visible_changes_in_min()

    def get_updated(self, **kwargs):
        """Get a new Scale object with updated arguments.

        Args:
            **kwargs (dict): the new keyword values, when given these take precedence over the current ones.

        Returns:
            Scale: a new scale with updated values.
        """
        new_values = dict(vmin=self.vmin, vmax=self.vmax, use_min=self.use_min, use_max=self.use_max)
        new_values.update(**kwargs)
        return Scale(**new_values)

    def __eq__(self, other):
        if not isinstance(other, Scale):
            return NotImplemented

        for key, value in self.__dict__.items():
            if value != getattr(other, key):
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)


class Font(object):

    def __init__(self, family='sans-serif', size=14):
        """Information about the font to use

        Args:
            name: the name of the font to use
            size: the size of the font (> 0).
        """
        self.family = family
        self.size = size

        if family not in self.font_names():
            raise ValueError("The given font \"{}\" is not recognized.".format(family))
        if size < 1:
            raise ValueError("The size ({}) can not be smaller than 1".format(str(size)))

    def get_updated(self, **kwargs):
        """Get a new Font object with updated arguments.

        Args:
            **kwargs (dict): the new keyword values, when given these take precedence over the current ones.

        Returns:
            Font: a new Font with updated values.
        """
        new_values = dict(family=self.family, size=self.size)
        new_values.update(**kwargs)
        return Font(**new_values)

    @property
    def name(self):
        return self.family

    @classmethod
    def font_names(cls):
        """Get the name of supported fonts

        Returns:
            list of str: the name of the supported fonts and font families.
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            fonts = matplotlib.font_manager.get_fontconfig_fonts()
            names = [matplotlib.font_manager.FontProperties(fname=font_name).get_name() for font_name in fonts]
        return list(sorted(['sans-serif', 'serif', 'cursive', 'fantasy', 'monospace'])) + list(sorted(names))

    @classmethod
    def get_conversion_info(cls):
        return SimpleClassConversion(cls, cls._get_attribute_conversions())

    @classmethod
    def _get_attribute_conversions(cls):
        return {'family': StringConversion(),
                'size': IntConversion()}

    def __eq__(self, other):
        if not isinstance(other, Font):
            return NotImplemented

        for key, value in self.__dict__.items():
            if value != getattr(other, key):
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)


class DataInfo(object):

    def __init__(self, maps, directory=None):
        """A container for basic information about the volume maps we are viewing.

        Args:
            maps (dict): the dictionary with the maps to view, these maps can either be arrays with values or
                nibabel proxy images.
            directory (str): the directory from which the maps where loaded
        """
        self._maps = maps
        self.directory = directory
        self.map_info = {key: SingleMapInfo(key, value) for key, value in maps.items()}
        self.sorted_keys = list(sorted(maps.keys()))

    @property
    def maps(self):
        return DeferredActionDict(lambda k, v: self.map_info[k].data, self.map_info)

    @classmethod
    def from_dir(cls, directory):
        if directory is None:
            return cls({}, None)
        return cls(nifti.load_all_niftis(directory), directory)

    def get_file_name(self, map_name):
        """Get the file name of the given map

        Returns:
            None if the map could not be found on dir, else a string with the file path.
        """
        if not self.directory:
            return None

        items = list(glob.glob(self.directory + '/{}.nii*'.format(map_name)))
        if items:
            return items[0]

        return None

    def get_max_dimension(self, map_names=None):
        """Get the minimum of the maximum dimension index over the maps

        Args:
            map_names (list of str): if given we will only scan the given list of maps

        Returns:
            int: either, 0, 1, 2 as the maximum dimension index in the maps.
        """
        map_names = map_names or self._maps.keys()
        if not map_names:
            raise ValueError('No maps to search in.')
        return min(self.map_info[map_name].max_dimension() for map_name in map_names)

    def get_max_slice_index(self, dimension, map_names=None):
        """Get the maximum slice index in the given map on the given dimension.

        Args:
            dimension (int): the dimension we want the slice index of (maximum 3)
            map_names (list of str): if given we will only scan the given list of maps

        Returns:
            int: the minimum of the maximum slice indices over the given maps in the given dimension.
        """
        map_names = map_names or self._maps.keys()
        max_dimension = self.get_max_dimension(map_names)
        if not map_names:
            raise ValueError('No maps to search in.')
        if dimension > max_dimension:
            raise ValueError('Dimension can not exceed {}.'.format(max_dimension))
        return min(self.map_info[map_name].max_slice_index(dimension) for map_name in map_names)

    def get_max_volume_index(self, map_names=None):
        """Get the maximum volume index in the given maps.

        In contrast to the max dimension and max slice index functions, this gives the maximum over all the
        images. This since handling different volumes is implemented in the viewer.

        Args:
            map_names (list of str): if given we will only scan the given list of maps

        Returns:
            int: the maximum volume index in the given list of maps. Starts from 0.
        """
        map_names = map_names or self._maps.keys()
        if not map_names:
            raise ValueError('No maps to search in.')
        return max(self.map_info[map_name].max_volume_index() for map_name in map_names)

    def get_index_first_non_zero_slice(self, dimension, map_names=None):
        """Get the index of the first non zero slice in the maps.

        Args:
            dimension (int): the dimension to search in
            map_names (list of str): if given we will only scan the given list of maps

        Returns:
            int: the slice index with the first non zero values.
        """
        map_names = map_names or self._maps.keys()
        if not map_names:
            raise ValueError('No maps to search in.')
        for map_name in map_names:
            index = self.map_info[map_name].get_index_first_non_zero_slice(dimension)
            if index is not None:
                return index
        return 0

    def slice_has_data(self, dimension, slice_index, map_names=None):
        """Check if at least one of the maps has non zero numbers on the given slice.

        Args:
            dimension (int): the dimension to search in
            slice_index (int): the index of the slice in the given dimension
            map_names (list of str): if given we will only scan the given list of maps

        Returns:
            bool: true if at least on of the maps has data in the given slice
        """
        map_names = map_names or self._maps.keys()
        if not map_names:
            raise ValueError('No maps to search in.')
        for map_name in map_names:
            if self.map_info[map_name].slice_has_data(dimension, slice_index):
                return True
        return False

    def get_max_x_index(self, dimension, rotate=0, map_names=None):
        """Get the maximum x index supported over the images.

        In essence this gets the lowest x index found.

        Args:
            dimension (int): the dimension to search in
            rotate (int): the rotation factor by which we rotate the slices within the given dimension
            map_names (list of str): if given we will only scan the given list of maps

        Returns:
            int: the maximum x-index found.
        """
        map_names = map_names or self._maps.keys()
        if not map_names:
            raise ValueError('No maps to search in.')
        return min(self.map_info[map_name].get_max_x_index(dimension, rotate) for map_name in map_names)

    def get_max_y_index(self, dimension, rotate=0, map_names=None):
        """Get the maximum y index supported over the images.

        In essence this gets the lowest y index found.

        Args:
            dimension (int): the dimension to search in
            rotate (int): the rotation factor by which we rotate the slices within the given dimension
            map_names (list of str): if given we will only scan the given list of maps

        Returns:
            int: the maximum y-index found.
        """
        map_names = map_names or self._maps.keys()
        if not map_names:
            raise ValueError('No maps to search in.')
        return min(self.map_info[map_name].get_max_y_index(dimension, rotate) for map_name in map_names)

    def get_bounding_box(self, dimension, slice_index, volume_index, rotate, map_names=None):
        """Get the bounding box of the images.

        Args:
            dimension (int): the dimension to search in
            slice_index (int): the slice index in that dimension
            volume_index (int): the current volume index
            rotate (int): the angle by which to rotate the image before getting the bounding box
            map_names (list of str): if given we will only scan the given list of maps

        Returns:
            tuple of Point: two point designating first the upper left corner and second the lower right corner of the
                bounding box.
        """
        map_names = map_names or self._maps.keys()
        if not map_names:
            raise ValueError('No maps to search in.')
        bounding_boxes = [self.map_info[map_name].get_bounding_box(dimension, slice_index, volume_index, rotate)
                          for map_name in map_names]

        p0x = min([bbox[0].x for bbox in bounding_boxes])
        p0y = min([bbox[0].y for bbox in bounding_boxes])
        p1x = max([bbox[1].x for bbox in bounding_boxes])
        p1y = max([bbox[1].y for bbox in bounding_boxes])

        return Point(p0x, p0y), Point(p1x, p1y)


class SingleMapInfo(object):

    def __init__(self, map_name, data):
        """Holds information about a single map.

        Args:
            map_name (str): the name of the map
            data (ndarray or :class:`nibabel.nifti1.Nifti1Image`): the value of the map or the proxy to it
        """
        self.map_name = map_name
        self._data = data
        self.shape = self._data.shape

    @property
    def data(self):
        if isinstance(self._data, nibabel.nifti1.Nifti1Image):
            return self._data.get_data()
        return self._data

    def max_dimension(self):
        """Get the maximum dimension index in this map.

        The maximum value returned by this method is 2 and the minimum is 0.

        Returns:
            int: in the range 0, 1, 2
        """
        return min(len(self.shape), 3) - 1

    def max_slice_index(self, dimension):
        """Get the maximum slice index on the given dimension.

        Args:
            dimension (int): the dimension we want the slice index of (maximum 3)

        Returns:
            int: the maximum slice index in the given dimension.
        """
        return self.shape[dimension] - 1

    def slice_has_data(self, dimension, slice_index):
        """Check if this map has non zero values in the given slice index.

        Args:
            dimension (int): the dimension we want the slice index of (maximum 3)
            slice_index (int): the slice index to look in

        Returns:
            int: the maximum slice index in the given dimension.
        """
        slice_indexing = [slice(None)] * (self.max_dimension() + 1)
        slice_indexing[dimension] = slice_index
        return np.count_nonzero(self.data[slice_indexing])

    def max_volume_index(self):
        """Get the maximum volume index in this map.

        The minimum is 0.

        Returns:
            int: the maximum volume index.
        """
        if len(self.shape) > 3:
            return self.shape[3] - 1
        return 0

    def get_index_first_non_zero_slice(self, dimension):
        """Get the index of the first non zero slice in this map.

        Args:
            dimension (int): the dimension to search in

        Returns:
            int: the slice index with the first non zero values.
        """
        slice_index = [slice(None)] * (self.max_dimension() + 1)

        if dimension > len(slice_index) - 1:
            raise ValueError('The given dimension {} is not supported.'.format(dimension))

        for index in range(self.shape[dimension]):
            slice_index[dimension] = index
            if np.count_nonzero(self.data[slice_index]) > 0:
                return index
        return 0

    def get_max_x_index(self, dimension, rotate=0):
        """Get the maximum x index.

        Args:
            dimension (int): the dimension to search in
            rotate (int): the value by which to rotate the slices in the given dimension

        Returns:
            int: the maximum x index
        """
        shape = list(self.shape)[0:3]
        del shape[dimension]
        if rotate // 90 % 2 == 0:
            return max(0, shape[1] - 1)
        return max(0, shape[0] - 1)

    def get_max_y_index(self, dimension, rotate=0):
        """Get the maximum y index.

        Args:
            dimension (int): the dimension to search in
            rotate (int): the value by which to rotate the slices in the given dimension

        Returns:
            int: the maximum y index
        """
        shape = list(self.shape)[0:3]
        del shape[dimension]
        if rotate // 90 % 2 == 0:
            return max(0, shape[0] - 1)
        return max(0, shape[1] - 1)

    def get_size_in_dimension(self, dimension, rotate=0):
        """Get the shape of the 2d view on the data in the given dimension.

        This basically returns a pair of (max_x, max_y).

        Args:
            dimension (int): the dimension to search in
            rotate (int): the value by which to rotate the slices in the given dimension

        Returns:
            tuple: (max_x, max_y)
        """
        return self.get_max_x_index(dimension, rotate), self.get_max_y_index(dimension, rotate)

    def get_bounding_box(self, dimension, slice_index, volume_index, rotate):
        """Get the bounding box of this map when displayed using the given indicing.

        Args:
            dimension (int): the dimension to search in
            slice_index (int): the slice index in that dimension
            volume_index (int): the current volume index
            rotate (int): the angle by which to rotate the image before getting the bounding box

        Returns:
            tuple of Point: two point designating first the upper left corner and second the lower right corner of the
                bounding box.
        """
        def bbox(image):
            rows = np.any(image, axis=1)
            cols = np.any(image, axis=0)

            rows_where = np.where(rows)

            if np.size(rows_where):
                row_min, row_max = np.where(rows)[0][[0, -1]]
                column_min, column_max = np.where(cols)[0][[0, -1]]

                return row_min, row_max, column_min, column_max
            return 0, image.shape[0]-1, 0, image.shape[1]-1

        slice_indexing = [slice(None)] * (self.max_dimension() + 1)
        slice_indexing[dimension] = slice_index

        image = self.data[slice_indexing]

        if len(image.shape) > 2:
            if image.shape[2] > 1:
                image = image[..., volume_index]
            else:
                image = image[..., 0]

        if rotate:
            image = np.rot90(image, rotate // 90)

        row_min, row_max, column_min, column_max = bbox(image)
        return Point(column_min, row_min), Point(column_max, row_max)


def get_available_interpolations():
    """The available interpolations for either the general map plot config or the map specifics.

    Do not call these for outside use, rather, consult the class method of the specific config you want to change.

    Returns:
        list of str: the list of available interpolations.
    """
    return ['none', 'nearest', 'bilinear', 'bicubic', 'spline16',
            'spline36', 'hanning', 'hamming', 'hermite', 'kaiser', 'quadric',
            'catrom', 'gaussian', 'bessel', 'mitchell', 'sinc', 'lanczos']


def get_available_colormaps():
    """The available colormaps for either the general map plot config or the map specifics.

    Do not call these for outside use, rather, consult the class method of the specific config you want to change.

    Returns:
        list of str: the list of available colormaps.
    """
    return sorted(matplotlib.cm.datad)



import numpy as np

from mdt.models.cascade import CascadeConfig

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class Tensor(CascadeConfig):

    name = 'Tensor (Cascade)'
    description = 'Cascade for Tensor.'
    models = ('BallStick_r1 (Cascade)',
              'Tensor')
    inits = {'Tensor': [('Tensor.theta', 'Stick.theta'),
                        ('Tensor.phi', 'Stick.phi')]}


class TensorFixed(CascadeConfig):

    name = 'Tensor (Cascade|fixed)'
    description = 'Cascade for Tensor with fixed angles.'
    models = ('BallStick_r1 (Cascade)',
              'Tensor')
    fixes = {'Tensor': [('Tensor.theta', 'Stick.theta'),
                        ('Tensor.phi', 'Stick.phi')]}


class TensorS0(CascadeConfig):

    name = 'Tensor (Cascade|S0)'
    description = 'Cascade for Tensor initialized with only an S0 fit.'
    models = ('S0',
              'Tensor')


class TensorExVivo(Tensor):

    name = 'Tensor-ExVivo (Cascade)'
    description = 'Cascade for Tensor with ex vivo defaults.'
    models = ('BallStick_r1-ExVivo (Cascade)',
              'Tensor-ExVivo')



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
"""Mask the background using the given mask.

This function multiplies a given volume (or list of volumes) with a binary mask.

Please note that this changes the input files (changes are in-place).
"""
import argparse
import glob
import os
import mdt
from argcomplete.completers import FilesCompleter
from mdt.shell_utils import BasicShellApplication
import textwrap

__author__ = 'Robbert Harms'
__date__ = "2015-08-18"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ApplyMask(BasicShellApplication):

    def _get_arg_parser(self):
        description = textwrap.dedent(__doc__)
        description += mdt.shell_utils.get_citation_message()

        epilog = textwrap.dedent("""
            Examples of use:
                mdt-apply-mask data.nii.gz -m roi_mask_0_50.nii.gz
                mdt-apply-mask *.nii.gz -m my_mask.nii.gz
        """)

        parser = argparse.ArgumentParser(description=description, epilog=epilog,
                                         formatter_class=argparse.RawTextHelpFormatter)

        parser.add_argument('input_files', metavar='input_files', nargs="+", type=str,
                            help="The input images to use")

        parser.add_argument('-m', '--mask', required=True,
                            action=mdt.shell_utils.get_argparse_extension_checker(['.nii', '.nii.gz', '.hdr', '.img']),
                            help='the (brain) mask to use').completer = FilesCompleter(['nii', 'gz', 'hdr', 'img'],
                                                                                       directories=False)
        return parser

    def run(self, args):
        mask = mdt.load_brain_mask(os.path.realpath(args.mask))

        file_names = []
        for file in args.input_files:
            file_names.extend(glob.glob(file))

        for file in file_names:
            mdt.apply_mask_to_file(file, mask)


if __name__ == '__main__':
    ApplyMask().start()



import six
import numpy as np

__author__ = 'Robbert Harms'
__date__ = "2015-08-25"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


def autodetect_brain_mask_loader(data_source):
    """A function to get a brain mask loader using the given data source.

    This tries to do auto detecting for the following data sources:

        - :class:`BrainMaskLoader`
        - strings (filenames)
        - ndarray (3d containing the mask)

    Args:
        data_source: the data source from which to get a brain_mask loader

    Returns:
        BrainMaskLoader: a brain_mask loader instance.
    """
    if isinstance(data_source, BrainMaskLoader):
        return data_source
    elif isinstance(data_source, six.string_types):
        return BrainMaskFromFileLoader(data_source)
    elif isinstance(data_source, np.ndarray):
        return BrainMaskFromArray(data_source)
    raise ValueError('The given data source could not be recognized.')


class BrainMaskLoader(object):
    """Interface for loading brain_masks from different sources."""

    def get_data(self):
        """The public method used to get an instance of a brain mask.

        Returns:
            ndarray: 3d ndarray containing the brain mask
        """


class BrainMaskFromFileLoader(BrainMaskLoader):

    def __init__(self, filename):
        """Loads a brain mask from the given filename.

        Args:
            filename (str): the filename to use the brain mask from.
        """
        self._filename = filename
        self._brain_mask = None
        self._header = None

    def get_data(self):
        if self._brain_mask is None:
            from mdt.nifti import load_nifti
            self._brain_mask = load_nifti(self._filename).get_data() > 0
        return self._brain_mask


class BrainMaskFromArray(BrainMaskLoader):

    def __init__(self, mask_data):
        """Adapter for returning an already loaded brain mask.

        Args:
            ndarray (ndarray): the brain mask data (3d matrix)
        """
        self._mask_data = mask_data

    def get_data(self):
        return self._mask_data > 0



from mdt.models.composite import DMRICompositeModelConfig

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CHARMED_r1(DMRICompositeModelConfig):

    ex_vivo_suitable = False
    description = 'The CHARMED model with 1 restricted compartments'

    model_expression = '''
        S0 * ( (Weight(w_hin0) * Tensor) +
               (Weight(w_res0) * CHARMEDRestricted(CHARMEDRestricted0))
               )
    '''

    lower_bounds = {'Tensor.d': 1e-9,
                    'Tensor.dperp0': 0.3e-9,
                    'Tensor.dperp1': 0.3e-9,
                    'CHARMEDRestricted0.d': 0.3e-9}

    upper_bounds = {'Tensor.d': 5e-9,
                    'Tensor.dperp0': 5e-9,
                    'Tensor.dperp1': 3e-9,
                    'CHARMEDRestricted0.d': 3e-9}

    inits = {'Tensor.d': 1.2e-9,
             'Tensor.dperp0': 0.5e-9,
             'Tensor.dperp1': 0.5e-9,
             'CHARMEDRestricted0.d': 1e-9}

    post_optimization_modifiers = [
        ('FR', lambda results: 1 - results['w_hin0.w'])
    ]


class CHARMED_r2(DMRICompositeModelConfig):

    ex_vivo_suitable = False
    description = 'The CHARMED model with 2 restricted compartments'

    model_expression = '''
        S0 * ( (Weight(w_hin0) * Tensor) +
               (Weight(w_res0) * CHARMEDRestricted(CHARMEDRestricted0)) +
               (Weight(w_res1) * CHARMEDRestricted(CHARMEDRestricted1)) )
    '''

    lower_bounds = {'Tensor.d': 1e-9,
                    'Tensor.dperp0': 0.3e-9,
                    'Tensor.dperp1': 0.3e-9,
                    'CHARMEDRestricted0.d': 0.3e-9,
                    'CHARMEDRestricted1.d': 0.3e-9}

    upper_bounds = {'Tensor.d': 5e-9,
                    'Tensor.dperp0': 5e-9,
                    'Tensor.dperp1': 3e-9,
                    'CHARMEDRestricted0.d': 3e-9,
                    'CHARMEDRestricted1.d': 3e-9}

    inits = {'Tensor.d': 1.2e-9,
             'Tensor.dperp0': 0.5e-9,
             'Tensor.dperp1': 0.5e-9,
             'CHARMEDRestricted0.d': 1e-9,
             'CHARMEDRestricted1.d': 1e-9}

    post_optimization_modifiers = [
        ('FR', lambda results: 1 - results['w_hin0.w'])
    ]


class CHARMED_r3(DMRICompositeModelConfig):

    ex_vivo_suitable = False
    description = 'The standard CHARMED model with 3 restricted compartments'

    model_expression = '''
        S0 * ( (Weight(w_hin0) * Tensor) +
               (Weight(w_res0) * CHARMEDRestricted(CHARMEDRestricted0)) +
               (Weight(w_res1) * CHARMEDRestricted(CHARMEDRestricted1)) +
               (Weight(w_res2) * CHARMEDRestricted(CHARMEDRestricted2)) )
    '''

    lower_bounds = {'Tensor.d': 1e-9,
                    'Tensor.dperp0': 0.3e-9,
                    'Tensor.dperp1': 0.3e-9,
                    'CHARMEDRestricted0.d': 0.3e-9,
                    'CHARMEDRestricted1.d': 0.3e-9,
                    'CHARMEDRestricted2.d': 0.3e-9}

    upper_bounds = {'Tensor.d': 5e-9,
                    'Tensor.dperp0': 5e-9,
                    'Tensor.dperp1': 3e-9,
                    'CHARMEDRestricted0.d': 3e-9,
                    'CHARMEDRestricted1.d': 3e-9,
                    'CHARMEDRestricted2.d': 3e-9}

    inits = {'Tensor.d': 1.2e-9,
             'Tensor.dperp0': 0.5e-9,
             'Tensor.dperp1': 0.5e-9,
             'CHARMEDRestricted0.d': 1e-9,
             'CHARMEDRestricted1.d': 1e-9,
             'CHARMEDRestricted2.d': 1e-9,
             'w_res2.w': 0}

    post_optimization_modifiers = [
        ('FR', lambda results: 1 - results['w_hin0.w'])
    ]



"""Definitions of the model data parameters.

These are meant for model specific data that the model needs to work. You can of course inline these variables in
the code for one of the models (which is faster), but this way lets the user change the specifics of the
model by changing the data in the model data parameters.

Please choose the parameter type for a model and parameter carefully since the type signifies how the parameter and
its data are handled during model construction.

"""

import numpy as np
from mdt.models.parameters import ModelDataParameterConfig

__author__ = 'Robbert Harms'
__date__ = "2015-12-12"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


# charmed default, used in GDRCylindersFixed model
class gamma_radii(ModelDataParameterConfig):

    data_type = 'global const mot_float_type* const'
    value = 1e-6 * np.array([1.5, 2.5, 3.5, 4.5, 5.5, 6.5])


class gamma_cyl_weights(ModelDataParameterConfig):

    data_type = 'global const mot_float_type* const'
    value = np.array([0.0211847200855742, 0.107169623942214,
                      0.194400551313197, 0.266676876170322,
                      0.214921653661151, 0.195646574827541])


class nmr_gamma_cyl_weights(ModelDataParameterConfig):

    data_type = 'int'
    value = 6



import collections
import copy

__author__ = 'Robbert Harms'
__date__ = "2016-11-10"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class DeferredActionDict(collections.MutableMapping):

    def __init__(self, func, items, memoize=True):
        """Applies the given function on the given items at moment of request.

        On the moment one of the keys of this dict class is requested we apply the given function on the given items
        and return the result of that function. The advantage of this class is that it defers an expensive operation
        until it is needed.

        Args:
            func (function): the callback function to apply on the given items at request, with signature:

                .. code-block: python

                    def callback(key, value)

            items (dict): the items on which we operate
            memoize (boolean): if true we memorize the function output internally. If False we apply the given function
                on every request.
        """
        self._func = func
        self._items = copy.copy(items)
        self._memoize = memoize
        self._memoized = {}

    def __delitem__(self, key):
        del self._items[key]
        if key in self._memoized:
            del self._memoized[key]

    def __getitem__(self, key):
        if not self._memoize:
            return self._func(key, self._items[key])

        if key not in self._memoized:
            self._memoized[key] = self._func(key, self._items[key])
        return self._memoized[key]

    def __contains__(self, key):
        try:
            self._items[key]
        except KeyError:
            return False
        else:
            return True

    def __iter__(self):
        for key in self._items.keys():
            yield key

    def __len__(self):
        return len(self._items)

    def __setitem__(self, key, value):
        self._memoized[key] = value


class DeferredFunctionDict(collections.MutableMapping):

    def __init__(self, items, memoize=True):
        """The items should contain a list of functions that we apply at the moment of request.

        On the moment one of the keys of this dict class is requested we apply the function stored in the items dict
        for that key and return the result of that function.
        The advantage of this class is that it defers an expensive operation until it is needed.

        Args:
            items (dict): the items on which we operate, each value should contain a function with no parameters
                that we run to return the results.
            memoize (boolean): if true we memorize the function output internally. If False we apply the item's function
                on every request.
        """
        self._items = copy.copy(items)
        self._memoize = memoize
        self._memoized = {}

    def __delitem__(self, key):
        del self._items[key]
        if key in self._memoized:
            del self._memoized[key]

    def __getitem__(self, key):
        if not self._memoize:
            return self._items[key]()

        if key not in self._memoized:
            self._memoized[key] = self._items[key]()
        return self._memoized[key]

    def __contains__(self, key):
        try:
            self._items[key]
        except KeyError:
            return False
        else:
            return True

    def __iter__(self):
        for key in self._items.keys():
            yield key

    def __len__(self):
        return len(self._items)

    def __setitem__(self, key, value):
        self._memoized[key] = value


class DeferredActionTuple(collections.Sequence):

    def __init__(self, func, items, memoize=True):
        """Applies the given function on the given items at moment of request.

        On the moment one of the elements is requested we apply the given function on the given items
        and return the result of that function. The advantage of this class is that it defers an expensive operation
        until it is needed.

        Args:
            func (function): the callback function to apply on the given items at request, with signature:

                .. code-block: python

                    def callback(index, value)

            items (list, tuple): the items on which we operate
            memoize (boolean): if true we memorize the function output internally. If False we apply the given function
                on every request.
        """
        self._func = func
        self._items = copy.copy(items)
        self._memoize = memoize
        self._memoized = {}

    def __getitem__(self, index):
        if not self._memoize:
            return self._func(index, self._items[index])

        if index not in self._memoized:
            self._memoized[index] = self._func(index, self._items[index])
        return self._memoized[index]

    def __len__(self):
        return len(self._items)
